#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.storage import *
from time import sleep


class TestStorageBucket:
    def test_put_file(self, exo, bucket, tmp_path):
        bucket = Bucket(exo.storage, {}, bucket())

        f_a = tmp_path / "file_a"
        f_a.write_text("a")

        f = bucket.put_file(str(f_a), metadata={"k": "v"})
        assert f.path == "file_a"
        res = exo.storage.boto.get_object(Bucket=bucket.name, Key="file_a")
        assert res["Body"].read() == b"a"
        assert res["Metadata"] == {"k": "v"}

        f = bucket.put_file(str(f_a), "file_b")
        assert f.path == "file_b"
        res = exo.storage.boto.get_object(Bucket=bucket.name, Key="file_b")

        exo.storage.boto.delete_object(Bucket=bucket.name, Key="file_a")
        exo.storage.boto.delete_object(Bucket=bucket.name, Key="file_b")

    def test_list_files(self, exo, bucket, tmp_path):
        bucket = Bucket(exo.storage, {}, bucket())

        f_a = tmp_path / "file_a"
        f_a.write_text("a")
        f_b = tmp_path / "file_b"
        f_b.write_text("b")

        exo.storage.boto.upload_file(Bucket=bucket.name, Filename=str(f_a), Key="a/a")
        exo.storage.boto.upload_file(Bucket=bucket.name, Filename=str(f_b), Key="b")

        files = list(bucket.list_files())
        assert len(files) == 2
        assert files[0].path == "a/a"
        assert files[1].path == "b"

        files = list(bucket.list_files(prefix="a"))
        assert len(files) == 1
        assert files[0].path == "a/a"

        exo.storage.boto.delete_object(Bucket=bucket.name, Key="a/a")
        exo.storage.boto.delete_object(Bucket=bucket.name, Key="b")

    def test_get_file(self, exo, bucket, tmp_path):
        bucket = Bucket(exo.storage, {}, bucket())

        f_a = tmp_path / "file_a"
        f_a.write_text("a")

        exo.storage.boto.upload_file(Bucket=bucket.name, Filename=str(f_a), Key="a")

        file = bucket.get_file("a")
        assert file.path == "a"
        assert file.bucket.name == bucket.name

        exo.storage.boto.delete_object(Bucket=bucket.name, Key="a")

    def test_delete_files(self, exo, bucket, tmp_path):
        bucket = Bucket(exo.storage, {}, bucket())

        f_a = tmp_path / "file_a"
        f_a.write_text("a")
        f_b = tmp_path / "file_b"
        f_b.write_text("b")

        exo.storage.boto.upload_file(Bucket=bucket.name, Filename=str(f_a), Key="a/a")
        exo.storage.boto.upload_file(Bucket=bucket.name, Filename=str(f_b), Key="b")

        res = exo.storage.boto.list_objects(Bucket=bucket.name)
        assert len(res["Contents"]) == 2

        bucket.delete_files(prefix="a")

        res = exo.storage.boto.list_objects(Bucket=bucket.name)
        assert len(res["Contents"]) == 1
        assert res["Contents"][0]["Key"] == "b"

        exo.storage.boto.delete_object(Bucket=bucket.name, Key="b")

    def test_delete(self, exo, bucket):
        bucket = Bucket(exo.storage, {}, bucket(teardown=False))
        bucket_name = bucket.name

        bucket.delete()
        assert bucket.name == None

        res = exo.storage.boto.list_buckets()
        for i in res.get("Buckets", []):
            assert i["Name"] != bucket_name

    def test_set_acl(self, exo, bucket):
        bucket = Bucket(exo.storage, {}, bucket())

        bucket.set_acl(acl="public-read")
        res = exo.storage.boto.get_bucket_acl(Bucket=bucket.name)
        for i in res["Grants"]:
            if i["Permission"] == "READ":
                assert i["Grantee"]["Type"] == "Group"
                assert i["Grantee"]["URI"] == ACL_ALL_USERS

        bucket_acp = AccessControlPolicy._from_s3(res)
        bucket_acp.read = "ALL_USERS"
        bucket_acp.write = "AUTHENTICATED_USERS"
        bucket_acp.read_acp = "alice@example.net"
        bucket_acp.write_acp = "bob@example.net"
        bucket.set_acl(acp=bucket_acp)
        res = exo.storage.boto.get_bucket_acl(Bucket=bucket.name)
        for i in res["Grants"]:
            if i["Permission"] == "READ":
                assert i["Grantee"]["Type"] == "Group"
                assert i["Grantee"]["URI"] == ACL_ALL_USERS
            if i["Permission"] == "WRITE":
                assert i["Grantee"]["Type"] == "Group"
                assert i["Grantee"]["URI"] == ACL_AUTHENTICATED_USERS
            if i["Permission"] == "READ_ACP":
                assert i["Grantee"]["Type"] == "CanonicalUser"
                assert i["Grantee"]["ID"] == "alice@example.net"
                assert i["Grantee"]["DisplayName"] == "alice@example.net"
            if i["Permission"] == "WRITE_ACP":
                assert i["Grantee"]["Type"] == "CanonicalUser"
                assert i["Grantee"]["ID"] == "bob@example.net"
                assert i["Grantee"]["DisplayName"] == "bob@example.net"

    def test_set_cors(self, exo, bucket):
        bucket = Bucket(exo.storage, {}, bucket())

        bucket.set_cors(
            [
                CORSRule(
                    {},
                    allowed_headers=["*"],
                    allowed_methods=["GET"],
                    allowed_origins=["https://example.net/"],
                    expose_headers=["X-Header1"],
                    max_age_seconds=3600,
                )
            ]
        )
        res = exo.storage.boto.get_bucket_cors(Bucket=bucket.name)
        assert len(res["CORSRules"]) == 1
        assert res["CORSRules"][0]["AllowedHeaders"] == ["*"]
        assert res["CORSRules"][0]["AllowedMethods"] == ["GET"]
        assert res["CORSRules"][0]["AllowedOrigins"] == ["https://example.net/"]
        assert res["CORSRules"][0]["ExposeHeaders"] == ["X-Header1"]
        assert res["CORSRules"][0]["MaxAgeSeconds"] == 3600

    def test_properties(self, exo, bucket, test_prefix):
        bucket = Bucket(exo.storage, {}, bucket(zone="ch-gva-2"))
        assert bucket.zone == "ch-gva-2"
        assert bucket.url.startswith(
            "https://sos-ch-gva-2.exo.io/{}".format(test_prefix)
        )

        exo.storage.boto.put_bucket_acl(
            Bucket=bucket.name,
            AccessControlPolicy={
                "Grants": [
                    {
                        "Grantee": {
                            "DisplayName": "alice@example.net",
                            "ID": "alice@example.net",
                            "Type": "CanonicalUser",
                        },
                        "Permission": "FULL_CONTROL",
                    },
                    {
                        "Grantee": {"Type": "Group", "URI": ACL_ALL_USERS},
                        "Permission": "READ",
                    },
                    {
                        "Grantee": {"Type": "Group", "URI": ACL_AUTHENTICATED_USERS},
                        "Permission": "WRITE",
                    },
                    {
                        "Grantee": {
                            "DisplayName": "bob@example.net",
                            "ID": "bob@example.net",
                            "Type": "CanonicalUser",
                        },
                        "Permission": "READ_ACP",
                    },
                    {
                        "Grantee": {
                            "DisplayName": "carl@example.net",
                            "ID": "carl@example.net",
                            "Type": "CanonicalUser",
                        },
                        "Permission": "WRITE_ACP",
                    },
                ]
            },
        )

        # Retrieving the bucket ACL policy immediately after having set it can trigger
        # a race condition at SOS level due to eventual consistency data store, so as
        # a workaround we wait for a bit before getting our ACL policy back.
        sleep(3)

        acp = bucket.acl
        assert acp.full_control == "alice@example.net"
        assert acp.read == "ALL_USERS"
        assert acp.write == "AUTHENTICATED_USERS"
        assert acp.read_acp == "bob@example.net"
        assert acp.write_acp == "carl@example.net"

        exo.storage.boto.put_bucket_cors(
            Bucket=bucket.name,
            CORSConfiguration={
                "CORSRules": [
                    {
                        "AllowedHeaders": ["*"],
                        "AllowedMethods": ["GET"],
                        "AllowedOrigins": ["https://example.net/"],
                        "ExposeHeaders": ["X-Header1"],
                        "MaxAgeSeconds": 3600,
                    }
                ]
            },
        )

        # Retrieving the CORS configuration immediately after having set it can trigger
        # a race condition at SOS level due to eventual consistency data store, so as
        # a workaround we wait for a bit before getting our CORS configuration back.
        sleep(3)

        cors = list(bucket.cors)
        assert len(cors) == 1
        assert cors[0].allowed_headers == ["*"]
        assert cors[0].allowed_methods == ["GET"]
        assert cors[0].allowed_origins == ["https://example.net/"]
        assert cors[0].expose_headers == ["X-Header1"]
        assert cors[0].max_age_seconds == 3600



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.compute import *


class TestComputeAntiAffinityGroup:
    def test_delete(self, exo, aag):
        anti_affinity_group = AntiAffinityGroup._from_cs(
            exo.compute, aag(teardown=False)
        )
        anti_affinity_group_name = anti_affinity_group.name

        anti_affinity_group.delete()
        assert anti_affinity_group.id is None

        res = exo.compute.cs.listAffinityGroups(
            name=anti_affinity_group_name, fetch_list=True
        )
        assert len(res) == 0



# -*- coding: utf-8 -*-

"""
This submodule represents the Exoscale Storage API.
"""

import attr
import boto3
import botocore
import logging
from . import API, Resource, APIException, ResourceNotFoundError
from os.path import basename

_DEFAULT_ZONE = "ch-gva-2"

_SUPPORTED_CANNED_ACLS = {
    "private",
    "public-read",
    "public-read-write",
    "authenticated-read",
    "bucket-owner-read",
    "bucket-owner-full-control",
}

ACL_ALL_USERS = "http://acs.amazonaws.com/groups/global/AllUsers"
ACL_AUTHENTICATED_USERS = "http://acs.amazonaws.com/groups/global/AuthenticatedUsers"


@attr.s
class CORSRule(Resource):
    """
    A Storage bucket CORS rule.

    Attributes:
        allowed_headers ([str]): list of allowed HTTP headers
        allowed_methods ([str]): list of allowed HTTP methods
        allowed_origins ([str]): list of allowed HTTP origins
        expose_headers ([str]): list of HTTP headers allowed to be exposed in response
        max_age_seconds (int): time in seconds that a browser can cache OPTIONS reponse
    """

    res = attr.ib(repr=False)
    allowed_headers = attr.ib(default=None, repr=False)
    allowed_methods = attr.ib(default=None, repr=False)
    allowed_origins = attr.ib(default=None, repr=False)
    expose_headers = attr.ib(default=None, repr=False)
    max_age_seconds = attr.ib(default=None, repr=False)

    @classmethod
    def _from_s3(cls, res):
        return cls(
            res,
            allowed_headers=res["AllowedHeaders"],
            allowed_methods=res["AllowedMethods"],
            allowed_origins=res["AllowedOrigins"],
            expose_headers=res.get("ExposeHeaders", None),
            max_age_seconds=res.get("MaxAgeSeconds", None),
        )

    def _to_s3(self):
        """
        Serialize a CORSRule class instance to the AWS S3 CORS rule format.

        References:
            * `AWS S3 CORS rule format`_

            .. _AWS S3 CORS rule format: https://docs.aws.amazon.com/AmazonS3/latest/API/RESTBucketPUTcors.html
        """

        cors_rule = {
            "AllowedHeaders": getattr(self, "allowed_headers", []),
            "AllowedMethods": getattr(self, "allowed_methods", []),
            "AllowedOrigins": getattr(self, "allowed_origins", []),
        }

        if self.expose_headers is not None:
            cors_rule["ExposeHeaders"] = self.expose_headers

        if self.max_age_seconds is not None:
            cors_rule["MaxAgeSeconds"] = self.max_age_seconds

        return cors_rule


@attr.s
class AccessControlPolicy(Resource):
    """
    A Storage Access Control Policy.

    Attributes:
        owner (str): entity owner of the resource
        full_control (str): full control grant
        read (str): read permission grant
        write (str): write permission grant
        read_acp (str): Access Control Policy read permission grant
        write_acp (str): Access Control Policy write permission grant
    """

    res = attr.ib(repr=False)
    owner = attr.ib()
    full_control = attr.ib(default=None, repr=False)
    read = attr.ib(default=None, repr=False)
    write = attr.ib(default=None, repr=False)
    read_acp = attr.ib(default=None, repr=False)
    write_acp = attr.ib(default=None, repr=False)

    @classmethod
    def _from_s3(cls, res):
        acp = cls(res, owner=res["Owner"]["DisplayName"])

        for i in res["Grants"]:
            if i["Permission"] == "FULL_CONTROL":
                acp.full_control = acp._grantee_from_s3(i["Grantee"])
            if i["Permission"] == "READ":
                acp.read = acp._grantee_from_s3(i["Grantee"])
            if i["Permission"] == "WRITE":
                acp.write = acp._grantee_from_s3(i["Grantee"])
            if i["Permission"] == "READ_ACP":
                acp.read_acp = acp._grantee_from_s3(i["Grantee"])
            if i["Permission"] == "WRITE_ACP":
                acp.write_acp = acp._grantee_from_s3(i["Grantee"])

        return acp

    def _to_s3(self):
        """
        Serialize an AccessControlPolicy class instance to the AWS S3 Access Control
        Policy format.

        References:
            * `boto3 AccessControlPolicy format`_

            .. _boto3 AccessControlPolicy format: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.put_bucket_acl
        """

        acp = {"Owner": {"ID": self.owner, "DisplayName": self.owner}, "Grants": []}

        if self.full_control is not None:
            acp["Grants"].append(
                {
                    "Grantee": self._grantee_to_s3(self.full_control),
                    "Permission": "FULL_CONTROL",
                }
            )
        if self.read is not None:
            acp["Grants"].append(
                {"Grantee": self._grantee_to_s3(self.read), "Permission": "READ"}
            )
        if self.write is not None:
            acp["Grants"].append(
                {"Grantee": self._grantee_to_s3(self.write), "Permission": "WRITE"}
            )
        if self.read_acp is not None:
            acp["Grants"].append(
                {
                    "Grantee": self._grantee_to_s3(self.read_acp),
                    "Permission": "READ_ACP",
                }
            )
        if self.write_acp is not None:
            acp["Grants"].append(
                {
                    "Grantee": self._grantee_to_s3(self.write_acp),
                    "Permission": "WRITE_ACP",
                }
            )

        return acp

    def _grantee_from_s3(self, grantee):
        """
        Convert an AWS S3 Access Control Policy grantee to AccessControlPolicy class
        format.
        """

        if grantee["Type"] == "CanonicalUser":
            return grantee["DisplayName"]
        else:
            if grantee["URI"] == ACL_ALL_USERS:
                return "ALL_USERS"
            elif grantee["URI"] == ACL_AUTHENTICATED_USERS:
                return "AUTHENTICATED_USERS"

        raise ValueError("unsupported grantee: {}".format(grantee))

    def _grantee_to_s3(self, grantee):
        """
        Convert an AccessControlPolicy class to AWS S3 Access Control Policy grantee
        format.

        References:
            * `boto3 AccessControlPolicy format`_

            .. _boto3 AccessControlPolicy format: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.put_bucket_acl
        """

        if grantee in {"ALL_USERS", "AUTHENTICATED_USERS"}:
            if grantee == "ALL_USERS":
                uri = ACL_ALL_USERS
            else:
                uri = ACL_AUTHENTICATED_USERS
            return {"Type": "Group", "URI": uri}
        else:
            return {"Type": "CanonicalUser", "ID": grantee, "DisplayName": grantee}


@attr.s
class BucketFile(Resource):
    """
    A file stored in Storage bucket.

    Attributes:
        path (str): the stored file path
        bucket (Bucket): the bucket the file is stored into
    """

    storage = attr.ib(repr=False)
    res = attr.ib(repr=False)
    path = attr.ib()
    bucket = attr.ib(repr=False)

    @property
    def content(self):
        """
        Stored file content.

        Returns:
            botocore.response.StreamingBody: the stored file content body

        Example:
            To read the content of a stored file::

                for f in mybucket.list_files():
                    print(f.content.read())

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.

        References:
            * `botocore.response API`_

            .. _botocore.response API: https://botocore.amazonaws.com/v1/documentation/api/latest/reference/response.html
        """

        try:
            res = self.storage.boto.get_object(Bucket=self.bucket.name, Key=self.path)
        except Exception as e:
            raise APIException(e)

        return res["Body"]

    @property
    def size(self):
        """
        Stored file size.

        Returns:
            int: the stored file size in bytes

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        if "Size" in self.res:
            return self.res["Size"]

        try:
            res = self.storage.boto.get_object(Bucket=self.bucket.name, Key=self.path)
        except Exception as e:
            raise APIException(e)

        return res["ContentLength"]

    @property
    def last_modification_date(self):
        """
        Stored file last modification date.

        Returns:
            datetime.datetime: the stored file last modification date

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.

        References:
            * `Python datetime module`_

            .. _Python datetime module: https://docs.python.org/3/library/datetime.html
        """

        if "LastModified" in self.res:
            return self.res["LastModified"]

        try:
            res = self.storage.boto.get_object(Bucket=self.bucket.name, Key=self.path)
        except Exception as e:
            raise APIException(e)

        return res["LastModified"]

    @property
    def acl(self):
        """
        Stored file Access Control List.

        Returns:
            AccessControlPolicy: the Access Control Policy applied to the stored file

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            res = self.storage.boto.get_object_acl(
                Bucket=self.bucket.name, Key=self.path
            )
        except Exception as e:
            raise APIException(e)

        res.pop("ResponseMetadata")
        return AccessControlPolicy._from_s3(res)

    @property
    def metadata(self):
        """
        Stored file metadata.

        Returns:
            dict: the stored file metadata

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            res = self.storage.boto.get_object(Bucket=self.bucket.name, Key=self.path)
        except Exception as e:
            raise APIException(e)

        return res["Metadata"]

    @property
    def url(self):
        """
        Stored file URL.

        Returns:
            str: the stored file URL

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        return self.bucket.url.rstrip("/") + "/" + self.path.lstrip("/")

    def set_acl(self, acl="", acp=None):
        """
        Set the stored file Access Control List.

        Parameters:
            acl (str): a canned ACL
            acp (AccessControlPolicy): an Access Control Policy

        Returns:
            None
        """

        if acl is None and acp is None:
            raise ValueError("either acl or acp must be specified")

        if acl != "" and acl not in _SUPPORTED_CANNED_ACLS:
            raise ValueError(
                "unsupported ACL; supported ACLs are: {}".format(
                    ",".join(_SUPPORTED_CANNED_ACLS)
                )
            )

        try:
            if acl != "":
                res = self.storage.boto.put_object_acl(
                    Bucket=self.bucket.name, Key=self.path, ACL=acl
                )
            else:
                res = self.storage.boto.put_object_acl(
                    Bucket=self.bucket.name,
                    Key=self.path,
                    AccessControlPolicy=acp._to_s3(),
                )
        except Exception as e:
            raise APIException(e)

    def delete(self):
        """
        Delete the stored file.

        Returns:
            None
        """

        try:
            res = self.storage.boto.delete_object(
                Bucket=self.bucket.name, Key=self.path
            )
        except Exception as e:
            raise APIException(e)

        self.storage = None
        self.res = None
        self.path = None
        self.bucket = None


@attr.s
class Bucket(Resource):
    """
    A Storage bucket.

    Attributes:
        name (str): the Storage bucket name
    """

    storage = attr.ib(repr=False)
    res = attr.ib(repr=False)
    name = attr.ib()

    @property
    def acl(self):
        """
        Storage bucket Access Control List.

        Returns:
            AccessControlPolicy: the Access Control Policy applied to the Storage bucket

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            res = self.storage.boto.get_bucket_acl(Bucket=self.name)
        except Exception as e:
            raise APIException(e)

        res.pop("ResponseMetadata")
        return AccessControlPolicy._from_s3(res)

    @property
    def cors(self):
        """
        Storage bucket Cross-Origin Resource Sharing configuration.

        Yields:
            CORSRule: the next CORS rule

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            res = self.storage.boto.get_bucket_cors(Bucket=self.name)
            for i in res["CORSRules"]:
                yield CORSRule._from_s3(i)
        except Exception as e:
            raise APIException(e)

    @property
    def zone(self):
        """
        Storage bucket zone.

        Returns:
            str: the Storage bucket zone

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        try:
            res = self.storage.boto.get_bucket_location(Bucket=self.name)
        except Exception as e:
            raise APIException(e)

        return res["LocationConstraint"]

    @property
    def url(self):
        """
        URL of the Storage bucket.

        Returns:
            str: the Storage bucket URL

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        return "https://sos-{zone}.exo.io/{bucket}".format(
            zone=self.zone, bucket=self.name
        )

    def put_file(self, src, dst=None, metadata=None, acl=None, transferConfig=None):
        """
        Store a file in the bucket.

        Parameters:
            src (str): the path to the source file
            dst (str): a path to which to store the file in the bucket
            metadata (dict): a dict of metadata to set to the file
            acl (str): a canned ACL to apply to the file
            transferConfig (boto3.s3.transfer.TransferConfig): a boto3 transfer
                configuration

        Returns:
            BucketFile: the file stored in the bucket
        """

        if dst is None:
            dst = basename(src)

        extraArgs = {}
        if metadata:
            extraArgs["Metadata"] = metadata
        if acl is not None:
            if acl not in _SUPPORTED_CANNED_ACLS:
                raise ValueError(
                    "unsupported ACL; supported ACLs are: {}".format(
                        ",".join(_SUPPORTED_CANNED_ACLS)
                    )
                )
            extraArgs["ACL"] = acl

        try:
            self.storage.boto.upload_file(
                Filename=src,
                Bucket=self.name,
                Key=dst,
                ExtraArgs=extraArgs if len(extraArgs) > 0 else None,
                Config=transferConfig,
            )
        except Exception as e:
            raise APIException(e)

        return BucketFile(self.storage, {}, path=dst, bucket=self)

    def list_files(self, prefix=""):
        """
        List files stored in the bucket.

        Parameters:
            prefix (str): a path prefix to restrict results to

        Yields:
            BucketFile: the next file stored in the bucket
        """

        try:
            paginator = self.storage.boto.get_paginator("list_objects_v2")
            pages = paginator.paginate(
                Bucket=self.name, Prefix=prefix, PaginationConfig={"PageSize": 100}
            )
            for page in pages:
                for i in page["Contents"]:
                    yield BucketFile(self.storage, i, i["Key"], self)
        except Exception as e:
            raise APIException(e)

    def get_file(self, path):
        """
        Get a file stored in the bucket.

        Parameters:
            path (str): the path to the file stored in the bucket

        Returns:
            BucketFile: file stored in the bucket
        """

        try:
            res = self.storage.boto.get_object(Bucket=self.name, Key=path)
        except Exception as e:
            raise APIException(e)

        res.pop("ResponseMetadata")
        return BucketFile(self.storage, res, path=path, bucket=self)

    def delete_files(self, prefix=""):
        """
        Delete files stored in the bucket.

        Parameters:
            prefix (str): a path prefix to restrict files deletion to

        Returns:
            None
        """

        _MAX_DELETE_BATCH = 1000

        batches = []
        batch = []
        n = 0
        # Iterate over the list of files to delete, and pack batches of
        # _MAX_DELETE_BATCH files to be executed once we've finished listing
        for f in self.list_files(prefix):
            batch.append({"Key": f.path})
            n = n + 1

            # We reached maximum batch capacity, queue it and move on to the rest of
            # the list
            if n >= _MAX_DELETE_BATCH:
                batches.append(batch)
                batch = []
                n = 0

        # Include the remaining non-batched files
        if len(batch) > 0:
            batches.append(batch)

        # Perform objects batch deletions
        try:
            for ob in batches:
                res = self.storage.boto.delete_objects(
                    Bucket=self.name, Delete={"Objects": ob}
                )
        except Exception as e:
            raise APIException(e)

    def set_acl(self, acl="", acp=None):
        """
        Set the Storage bucket Access Control List.

        Parameters:
            acl (str): a canned ACL
            acp (AccessControlPolicy): an Access Control Policy

        Returns:
            None
        """

        if acl is None and acp is None:
            raise ValueError("either acl or acp must be specified")

        if acl != "" and acl not in _SUPPORTED_CANNED_ACLS:
            raise ValueError(
                "unsupported ACL; supported ACLs are: {}".format(
                    ",".join(_SUPPORTED_CANNED_ACLS)
                )
            )

        try:
            if acl != "":
                res = self.storage.boto.put_bucket_acl(Bucket=self.name, ACL=acl)
            else:
                res = self.storage.boto.put_bucket_acl(
                    Bucket=self.name, AccessControlPolicy=acp._to_s3()
                )
        except Exception as e:
            raise APIException(e)

    def set_cors(self, rules):
        """
        Set the Storage bucket Cross-Origin Resource Sharing configuration.

        Parameters:
            rules ([CORSRule]): a list of CORS rules

        Returns:
            None
        """

        try:
            res = self.storage.boto.put_bucket_cors(
                Bucket=self.name,
                CORSConfiguration={"CORSRules": list(r._to_s3() for r in rules)},
            )
        except Exception as e:
            raise APIException(e)

    def delete(self):
        """
        Delete the Storage bucket.

        Returns:
            None
        """

        try:
            res = self.storage.boto.delete_bucket(Bucket=self.name)
        except Exception as e:
            raise APIException(e)

        self.storage = None
        self.res = None
        self.name = None


class StorageAPI(API):
    """
    An Exoscale Object Storage API client.

    Parameters:
        key (str): the Storage API key
        secret (str): the Storage API secret
        endpoint (str): the Storage API endpoint
        zone (str): the Storage zone
        max_retries (int): the API HTTP session retry policy number of retries to allow
        trace (bool): API request/response tracing flag
    """

    def __init__(
        self, key, secret, endpoint=None, zone=None, max_retries=None, trace=False
    ):
        self.zone = _DEFAULT_ZONE if zone is None else zone
        endpoint = (
            "https://sos-{}.exo.io".format(self.zone) if endpoint is None else endpoint
        )
        max_retries = 3 if max_retries is None else max_retries
        super().__init__(
            endpoint=endpoint,
            key=key,
            secret=secret,
            max_retries=max_retries,
            trace=trace,
        )

        if self.zone is None:
            raise ValueError("no storage zone specified")

        self.boto = boto3.client(
            "s3",
            region_name=self.zone,
            endpoint_url=self.endpoint,
            aws_access_key_id=key,
            aws_secret_access_key=secret,
            config=botocore.client.Config(
                user_agent="{} boto3/{} botocore/{}".format(
                    self.user_agent, boto3.__version__, botocore.__version__
                ),
                retries={"max_attempts": self.max_retries},
            ),
        )

        if trace:
            boto3.set_stream_logger(name="", level=logging.DEBUG)

    def __repr__(self):
        return "StorageAPI(endpoint='{}' zone='{}' key='{}')".format(
            self.endpoint, self.zone, self.key
        )

    def __str__(self):
        return self.__repr__()

    def create_bucket(self, name, zone=None, acl="private"):
        """
        Create a Storage bucket.

        Parameters:
            name (str): the Storage bucket name
            zone (str): a Storage bucket zone
            acl (str): a canned ACL to apply to the bucket

        Returns:
            Bucket: the Storage bucket created
        """

        if acl not in _SUPPORTED_CANNED_ACLS:
            raise ValueError(
                "unsupported ACL; supported ACLs are: {}".format(
                    ",".join(_SUPPORTED_CANNED_ACLS)
                )
            )

        try:
            res = self.boto.create_bucket(
                Bucket=name,
                CreateBucketConfiguration={
                    "LocationConstraint": self.zone if zone is None else zone
                },
                ACL=acl,
            )
        except Exception as e:
            raise APIException(e)

        return Bucket(self, {}, name=name)

    def list_buckets(self):
        """
        List Storage buckets.

        Yields:
            Bucket: the next Storage bucket
        """

        try:
            res = self.boto.list_buckets()
        except Exception as e:
            raise APIException(e)

        if "Buckets" in res:
            for bucket in res["Buckets"]:
                yield Bucket(self, bucket, name=bucket["Name"])

    def get_bucket(self, name):
        """
        Get a Storage bucket.

        Parameters:
            name (str): a Storage bucket name

        Returns:
            Bucket: a Storage bucket
        """

        for bucket in self.list_buckets():
            if bucket.name == name:
                return bucket

        raise ResourceNotFoundError



# -*- coding: utf-8 -*-

import attr
import os
import pathlib
import toml
from .api.compute import ComputeAPI
from .api.dns import DnsAPI
from .api.storage import StorageAPI
from .api.runstatus import RunstatusAPI
from .api.iam import IamAPI

_API_KEY_ENVVAR = "EXOSCALE_API_KEY"
_API_SECRET_ENVVAR = "EXOSCALE_API_SECRET"
_COMPUTE_API_ENDPOINT_ENVVAR = "EXOSCALE_COMPUTE_API_ENDPOINT"
_DNS_API_ENDPOINT_ENVVAR = "EXOSCALE_DNS_API_ENDPOINT"
_RUNSTATUS_API_ENDPOINT_ENVVAR = "EXOSCALE_RUNSTATUS_API_ENDPOINT"
_STORAGE_API_ENDPOINT_ENVVAR = "EXOSCALE_STORAGE_API_ENDPOINT"
_IAM_API_ENDPOINT_ENVVAR = "EXOSCALE_IAM_API_ENDPOINT"
_STORAGE_ZONE_ENVVAR = "EXOSCALE_STORAGE_ZONE"
_CONFIG_FILE_ENVVAR = "EXOSCALE_CONFIG_FILE"
_DEFAULT_CONFIG_FILE = os.path.join(pathlib.Path.home(), "config.toml")


class Exoscale:
    """
    An Exoscale API client.

    Parameters:
        config_file (str): path to a configuration file
        profile (str): a configuration profile name
        api_key (str): an Exoscale client API key
        api_secret (str): an Exoscale client API secret
        compute_api_endpoint (str): an alternative Exoscale Compute API endpoint
        dns_api_endpoint (str): an alternative Exoscale DNS API endpoint
        storage_api_endpoint (str): an alternative Exoscale Object Storage API endpoint
        iam_api_endpoint (str): an alternative Exoscale IAM API endpoint
        storage_zone (str): an Exoscale Storage zone
        runstatus_api_endpoint (str): an alternative Exoscale Runstatus API endpoint
        max_retries (int): the API HTTP session retry policy number of retries to allow
        trace (bool): API requests/responses tracing flag

    Attributes:
        api_key (str): the Exoscale client API key
        api_secret (str): the Exoscale client API secret
        compute (api.compute.ComputeAPI): the Exoscale Compute API client
        dns (api.compute.DnsAPI): the Exoscale DNS API client
        storage (api.storage.StorageAPI): the Exoscale Object Storage API client
        runstatus (api.runstatus.RunstatusAPI): the Exoscale Runstatus API client
        iam (api.iam.IamAPI): the Exoscale IAM API client
    """

    def __init__(
        self,
        config_file=None,
        profile=None,
        api_key=None,
        api_secret=None,
        compute_api_endpoint=None,
        dns_api_endpoint=None,
        storage_api_endpoint=None,
        storage_zone=None,
        runstatus_api_endpoint=None,
        iam_api_endpoint=None,
        max_retries=3,
        trace=False,
    ):
        # Load settings from a configuration file profile
        config_file = (
            config_file if config_file else os.getenv(_CONFIG_FILE_ENVVAR, None)
        )
        if config_file:
            with open(config_file) as f:
                self._config = toml.load(f)

        if not getattr(self, "_config", None) and os.path.exists(_DEFAULT_CONFIG_FILE):
            with open(_DEFAULT_CONFIG_FILE) as f:
                self._config = toml.load(f)

        if getattr(self, "_config", None) is not None:
            profile = self._get_profile(profile)
            for k in {"name", "api_key", "api_secret"}:
                if k not in profile:
                    raise ConfigurationError('profile missing "{}" key'.format(k))

            api_key = profile["api_key"] if not api_key else api_key
            api_secret = profile["api_secret"] if not api_secret else api_secret
            compute_api_endpoint = (
                profile.get("compute_api_endpoint", None)
                if not compute_api_endpoint
                else compute_api_endpoint
            )
            dns_api_endpoint = (
                profile.get("dns_api_endpoint", None)
                if not dns_api_endpoint
                else dns_api_endpoint
            )
            storage_api_endpoint = (
                profile.get("storage_api_endpoint", None)
                if not storage_api_endpoint
                else storage_api_endpoint
            )
            storage_zone = (
                profile.get("storage_zone", None) if not storage_zone else storage_zone
            )
            runstatus_api_endpoint = (
                profile.get("runstatus_api_endpoint", None)
                if not runstatus_api_endpoint
                else runstatus_api_endpoint
            )
            iam_api_endpoint = (
                profile.get("iam_api_endpoint", None)
                if not iam_api_endpoint
                else iam_api_endpoint
            )

        # Fallback: load settings from environment variables
        api_key = api_key if api_key else os.getenv(_API_KEY_ENVVAR, None)
        api_secret = api_secret if api_secret else os.getenv(_API_SECRET_ENVVAR, None)
        compute_api_endpoint = (
            compute_api_endpoint
            if compute_api_endpoint
            else os.getenv(_COMPUTE_API_ENDPOINT_ENVVAR, None)
        )
        dns_api_endpoint = (
            dns_api_endpoint
            if dns_api_endpoint
            else os.getenv(_DNS_API_ENDPOINT_ENVVAR, None)
        )
        storage_api_endpoint = (
            storage_api_endpoint
            if storage_api_endpoint
            else os.getenv(_STORAGE_API_ENDPOINT_ENVVAR, None)
        )
        storage_zone = (
            storage_zone if storage_zone else os.getenv(_STORAGE_ZONE_ENVVAR, None)
        )
        runstatus_api_endpoint = (
            runstatus_api_endpoint
            if runstatus_api_endpoint
            else os.getenv(_RUNSTATUS_API_ENDPOINT_ENVVAR, None)
        )
        iam_api_endpoint = (
            iam_api_endpoint
            if iam_api_endpoint
            else os.getenv(_IAM_API_ENDPOINT_ENVVAR, None)
        )

        if api_key is None or api_secret is None:
            raise ConfigurationError("missing API key/secret")

        self.api_key = api_key
        self.api_secret = api_secret

        kwargs = {
            "key": self.api_key,
            "secret": self.api_secret,
            "max_retries": max_retries,
            "trace": trace,
        }
        if compute_api_endpoint is not None:
            kwargs["endpoint"] = compute_api_endpoint
        self.compute = ComputeAPI(**kwargs)

        kwargs = {
            "key": self.api_key,
            "secret": self.api_secret,
            "max_retries": max_retries,
            "trace": trace,
        }
        if dns_api_endpoint is not None:
            kwargs["endpoint"] = dns_api_endpoint
        self.dns = DnsAPI(**kwargs)

        kwargs = {
            "key": self.api_key,
            "secret": self.api_secret,
            "max_retries": max_retries,
            "trace": trace,
        }
        if storage_api_endpoint is not None:
            kwargs["endpoint"] = storage_api_endpoint
        if storage_zone is not None:
            kwargs["zone"] = storage_zone
        self.storage = StorageAPI(**kwargs)

        kwargs = {
            "key": self.api_key,
            "secret": self.api_secret,
            "max_retries": max_retries,
            "trace": trace,
        }
        if runstatus_api_endpoint is not None:
            kwargs["endpoint"] = runstatus_api_endpoint
        self.runstatus = RunstatusAPI(**kwargs)

        kwargs = {
            "key": self.api_key,
            "secret": self.api_secret,
            "max_retries": max_retries,
            "trace": trace,
        }
        if iam_api_endpoint is not None:
            kwargs["endpoint"] = iam_api_endpoint
        self.iam = IamAPI(**kwargs)

    def _get_profile(self, profile=None):
        if "profiles" not in self._config or len(self._config["profiles"]) == 0:
            raise ConfigurationError("no profiles configured")

        if profile or "default_profile" in self._config:
            if "default_profile" in self._config and not profile:
                profile = self._config["default_profile"]

            for a in self._config["profiles"]:
                if a["name"] == profile:
                    return a

            raise ConfigurationError('profile "{}" not found'.format(profile))

        return self._config["profiles"][0]

    def __repr__(self):
        return "Exoscale(api_key='{}')".format(self.api_key)

    def __str__(self):
        return self.__repr__()


class ConfigurationError(Exception):
    """
    A generic configuration error.
    """

    reason = attr.ib()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.runstatus import *
from datetime import timezone


class TestRunstatusPageIncident:
    def test_add_event(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())
        test_incident_title = "Everything's on fire"
        test_incident_description = "It's fine ¯\\_(ツ)_/¯"
        test_incident_state = "major_outage"
        test_incident_status = "identified"
        test_incident_services = ["a", "b", "c"]

        for i in test_incident_services:
            exo.runstatus._post(
                url="/pages/{p}/services".format(p=page.name), json={"name": i}
            )

        exo.runstatus._post(
            url="/pages/{p}/incidents".format(p=page.name),
            json={
                "title": test_incident_title,
                "status_text": test_incident_description,
                "status": test_incident_status,
                "state": test_incident_state,
                "services": test_incident_services,
            },
        )

        res = exo.runstatus._get(url="/pages/{p}/incidents".format(p=page.name)).json()
        incident = Incident._from_rs(exo.runstatus, res["results"][0], page)

        incident.add_event(description="Hmm...", status="investigating")
        incident.add_event(description="Who did this?!", status="identified")
        incident.add_event(
            description="Should be better now",
            status="monitoring",
            state="degraded_performance",
        )

        res = exo.runstatus._get(
            url="/pages/{p}/incidents/{i}/events".format(p=page.name, i=incident.id)
        ).json()
        assert len(res["results"]) == 4
        assert res["results"][0]["text"] == "Should be better now"
        assert res["results"][0]["state"] == "degraded_performance"
        assert res["results"][0]["status"] == "monitoring"
        assert res["results"][1]["text"] == "Who did this?!"
        assert res["results"][1]["state"] == "major_outage"
        assert res["results"][1]["status"] == "identified"
        assert res["results"][2]["text"] == "Hmm..."
        assert res["results"][2]["state"] == "major_outage"
        assert res["results"][2]["status"] == "investigating"

    def test_update(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())
        test_incident_description = "It's fine ¯\\_(ツ)_/¯"
        test_incident_state = "major_outage"
        test_incident_status = "identified"
        test_incident_title = "Everything's on fire"
        test_incident_title_edited = "Everything's on fire (edited)"
        test_incident_service = "db1"
        test_incident_service_edited = "db2"

        for i in [test_incident_service, test_incident_service_edited]:
            exo.runstatus._post(
                url="/pages/{p}/services".format(p=page.name), json={"name": i}
            )

        exo.runstatus._post(
            url="/pages/{p}/incidents".format(p=page.name),
            json={
                "title": test_incident_title,
                "status_text": test_incident_description,
                "status": test_incident_status,
                "state": test_incident_state,
                "services": [test_incident_service],
            },
        )

        res = exo.runstatus._get(url="/pages/{p}/incidents".format(p=page.name)).json()
        incident = Incident._from_rs(exo.runstatus, res["results"][0], page)

        incident.update(
            title=test_incident_title_edited, services=[test_incident_service_edited]
        )

        res = exo.runstatus._get(
            url="/pages/{p}/incidents/{i}".format(p=page.name, i=incident.id)
        ).json()
        assert res["title"] == test_incident_title_edited
        assert incident.title == test_incident_title_edited
        assert res["services"] == [test_incident_service_edited]
        assert incident.services == [test_incident_service_edited]

    def test_close(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())
        test_incident_title = "Everything's on fire"
        test_incident_description = "It's fine ¯\\_(ツ)_/¯"
        test_incident_state = "major_outage"
        test_incident_status = "identified"
        test_incident_services = ["a", "b", "c"]

        for i in test_incident_services:
            exo.runstatus._post(
                url="/pages/{p}/services".format(p=page.name), json={"name": i}
            )

        exo.runstatus._post(
            url="/pages/{p}/incidents".format(p=page.name),
            json={
                "title": test_incident_title,
                "status_text": test_incident_description,
                "status": test_incident_status,
                "state": test_incident_state,
                "services": test_incident_services,
            },
        )

        res = exo.runstatus._get(url="/pages/{p}/incidents".format(p=page.name)).json()
        incident = Incident._from_rs(exo.runstatus, res["results"][0], page)

        incident.close(description="There, I fixed it")

        res = exo.runstatus._get(
            url="/pages/{p}/incidents/{i}/events".format(p=page.name, i=incident.id)
        ).json()
        assert len(res["results"]) == 2
        assert res["results"][0]["text"] == "There, I fixed it"
        assert res["results"][0]["state"] == "operational"
        assert res["results"][0]["status"] == "resolved"

    def test_properties(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())
        test_incident_title = "Everything's on fire"
        test_incident_description = "It's fine ¯\\_(ツ)_/¯"
        test_incident_state = "major_outage"
        test_incident_status = "identified"
        test_incident_services = ["a", "b", "c"]

        for i in test_incident_services:
            exo.runstatus._post(
                url="/pages/{p}/services".format(p=page.name), json={"name": i}
            )

        exo.runstatus._post(
            url="/pages/{p}/incidents".format(p=page.name),
            json={
                "title": test_incident_title,
                "status_text": test_incident_description,
                "status": test_incident_status,
                "state": test_incident_state,
                "services": test_incident_services,
            },
        )

        res = exo.runstatus._get(url="/pages/{p}/incidents".format(p=page.name)).json()
        incident = Incident._from_rs(exo.runstatus, res["results"][0], page)

        for i in [1, 2, 3]:
            exo.runstatus._post(
                url="/pages/{p}/incidents/{i}/events".format(
                    p=page.name, i=incident.id
                ),
                json={
                    "text": "Update #{}".format(i),
                    "status": incident.status,
                    "state": incident.state,
                },
            )

        incident_events = list(incident.events)
        assert len(incident_events) == 4
        assert incident_events[0].description == "Update #3"
        assert incident_events[1].description == "Update #2"
        assert incident_events[2].description == "Update #1"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.compute import *


class TestComputePrivateNetwork:
    def test_attach_instance(self, exo, privnet, instance):
        private_network = PrivateNetwork._from_cs(exo.compute, privnet())
        instance = Instance._from_cs(exo.compute, instance())

        private_network.attach_instance(instance)

        res = exo.compute.cs.listNics(
            virtualmachineid=instance.id, networkid=private_network.id, fetch_list=True
        )
        assert len(res) == 1
        assert res[0]["networkid"] == private_network.id
        assert res[0]["virtualmachineid"] == instance.id

        for nic in res:
            if nic["isdefault"]:
                continue
            exo.compute.cs.removeNicFromVirtualMachine(
                virtualmachineid=instance.id, nicid=nic["id"]
            )

    def test_detach_instance(self, exo, privnet, instance):
        private_network = PrivateNetwork._from_cs(exo.compute, privnet())
        instance = Instance._from_cs(exo.compute, instance())

        exo.compute.cs.addNicToVirtualMachine(
            virtualmachineid=instance.id, networkid=private_network.id
        )

        [res] = exo.compute.cs.listNics(
            virtualmachineid=instance.id, networkid=private_network.id, fetch_list=True
        )
        assert res["networkid"] == private_network.id
        assert res["virtualmachineid"] == instance.id

        private_network.detach_instance(instance)

        res = exo.compute.cs.listNics(
            virtualmachineid=instance.id, networkid=private_network.id, fetch_list=True
        )
        assert len(res) == 0

    def test_update(self, exo, privnet):
        private_network = PrivateNetwork._from_cs(
            exo.compute,
            privnet(start_ip="10.0.0.10", end_ip="10.0.0.50", netmask="255.255.255.0"),
        )
        name_edited = private_network.name + " (edited)"
        description_edited = private_network.description + " (edited)"
        start_ip_edited = "10.0.0.1"
        end_ip_edited = "10.0.0.100"
        netmask_edited = "255.0.0.0"

        private_network.update(
            name=name_edited,
            description=description_edited,
            start_ip=start_ip_edited,
            end_ip=end_ip_edited,
            netmask=netmask_edited,
        )

        res = exo.compute.cs.listNetworks(id=private_network.id, fetch_list=True)
        assert res[0]["name"] == name_edited
        assert private_network.name == name_edited
        assert res[0]["displaytext"] == description_edited
        assert private_network.description == description_edited
        assert res[0]["startip"] == start_ip_edited
        assert private_network.start_ip == start_ip_edited
        assert res[0]["endip"] == end_ip_edited
        assert private_network.end_ip == end_ip_edited
        assert res[0]["netmask"] == netmask_edited
        assert private_network.netmask == netmask_edited

    def test_delete(self, exo, privnet):
        private_network = PrivateNetwork._from_cs(exo.compute, privnet(teardown=False))
        private_network_id = private_network.id

        private_network.delete()
        assert private_network.id is None

        with pytest.raises(CloudStackApiException) as excinfo:
            res = exo.compute.cs.listNetworks(id=private_network_id, fetch_list=True)
            assert len(res) == 0
        assert excinfo.type == CloudStackApiException
        assert "does not exist" in excinfo.value.error["errortext"]

    def test_properties(self, exo, privnet, instance):
        private_network = PrivateNetwork._from_cs(exo.compute, privnet())
        instance = Instance._from_cs(exo.compute, instance())

        res = exo.compute.cs.addNicToVirtualMachine(
            virtualmachineid=instance.id, networkid=private_network.id
        )

        private_network_instances = list(private_network.instances)
        assert len(private_network_instances) == 1
        assert private_network_instances[0].name == instance.name

        for nic in res["virtualmachine"]["nic"]:
            if nic["isdefault"]:
                continue
            exo.compute.cs.removeNicFromVirtualMachine(
                virtualmachineid=instance.id, nicid=nic["id"]
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.iam import *


class TestIAMAPIKey:
    def test_revoke(self, exo, apikey):
        api_key = APIKey._from_cs(exo.iam, apikey(teardown=False))
        api_key_id = api_key.key

        api_key.revoke()
        assert api_key.name is None

        with pytest.raises(CloudStackApiException) as excinfo:
            _api_key = exo.iam.cs.getApiKey(key=api_key_id)
            assert _api_key is None
        assert excinfo.type == CloudStackApiException



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.compute import *


class TestComputeInstanceVolumeSnapshot:
    def test_revert(self, exo, instance):
        instance = Instance._from_cs(exo.compute, instance())

        res = exo.compute.cs.createSnapshot(volumeid=instance.volume_id)
        snapshot = InstanceVolumeSnapshot._from_cs(exo.compute, res["snapshot"])

        res = exo.compute.cs.stopVirtualMachine(id=instance.id)
        assert res["virtualmachine"]["state"].lower() in ["stopping", "stopped"]

        snapshot.revert()
        # We don't have a way to figure out if the snapshot revert has actually worked,
        # so we'll consider that no exception is an OK -- no news is good news!

    def test_delete(self, exo, instance):
        instance = Instance._from_cs(exo.compute, instance())

        res = exo.compute.cs.createSnapshot(volumeid=instance.volume_id)
        snapshot = InstanceVolumeSnapshot._from_cs(exo.compute, res["snapshot"])

        snapshot.delete()
        assert snapshot.id is None

        res = exo.compute.cs.listSnapshots(volumeid=instance.volume_id, fetch_list=True)
        assert len(res) == 0

    def test_properties(self, exo, instance):
        instance = Instance._from_cs(exo.compute, instance())

        res = exo.compute.cs.createSnapshot(volumeid=instance.volume_id)
        instance_volume_snapshot = InstanceVolumeSnapshot._from_cs(
            exo.compute, res["snapshot"]
        )
        assert instance_volume_snapshot.state == "exported"

    def test_export(self, exo, instance):
        instance = Instance._from_cs(exo.compute, instance())

        res = exo.compute.cs.createSnapshot(volumeid=instance.volume_id)
        snapshot = InstanceVolumeSnapshot._from_cs(exo.compute, res["snapshot"])
        assert snapshot.state == "exported"

        res = snapshot.export()
        assert snapshot.state == "exported"
        assert "presignedurl" in res



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api import ResourceNotFoundError
from exoscale.api.compute import *
from cs import CloudStackApiException
from datetime import datetime, timedelta
from time import sleep


class TestComputeInstancePool:
    def test_update(self, exo, zone, instance_pool):
        zone = Zone._from_cs(zone("ch-gva-2"))
        instance_pool = InstancePool._from_cs(
            exo.compute, instance_pool(zone_id=zone.id)
        )
        name_edited = instance_pool.name + "-edited"
        description_edited = instance_pool.description + " (edited)"

        instance_pool.update(name=name_edited, description=description_edited)

        [actual_instance_pool] = exo.compute.cs.getInstancePool(
            id=instance_pool.id, zoneid=instance_pool.zone.id, fetch_list=True
        )
        assert instance_pool.name == name_edited
        assert actual_instance_pool["name"] == name_edited
        assert instance_pool.description == description_edited
        assert actual_instance_pool["description"] == description_edited

    def test_scale(self, exo, zone, instance_pool):
        zone = Zone._from_cs(zone("ch-gva-2"))
        instance_pool = InstancePool._from_cs(
            exo.compute, instance_pool(zone_id=zone.id)
        )

        instance_pool.scale(2)

        [actual_instance_pool] = exo.compute.cs.getInstancePool(
            id=instance_pool.id, zoneid=instance_pool.zone.id, fetch_list=True
        )
        assert instance_pool.size == 2
        assert actual_instance_pool["size"] == 2

    def test_delete_no_wait(self, exo, zone, instance_pool):
        zone = Zone._from_cs(zone("ch-gva-2"))
        instance_pool = InstancePool._from_cs(
            exo.compute, instance_pool(zone_id=zone.id, teardown=False)
        )
        instance_pool_id = instance_pool.id
        instance_pool_zone_id = instance_pool.zone.id

        # Let's wait until the Instance Pool fixture state stabilizes before destroying
        # it, otherwise we'll likely run into a race condition
        t = datetime.now()
        while exo.compute.cs.getInstancePool(
            zoneid=zone.id, id=instance_pool.id, fetch_list=True
        )[0]["state"] == "scaling-up" and datetime.now() - t < timedelta(minutes=5):
            sleep(10)

        instance_pool.delete(wait=False)
        assert instance_pool.id is None

        # Wait a few seconds for the deletion process to kick in
        sleep(5)

        [actual_instance_pool] = exo.compute.cs.getInstancePool(
            id=instance_pool_id, zoneid=instance_pool_zone_id, fetch_list=True
        )
        assert actual_instance_pool["state"] == "destroying"

    def test_delete_wait(self, exo, zone, instance_pool):
        zone = Zone._from_cs(zone("ch-gva-2"))
        instance_pool = InstancePool._from_cs(
            exo.compute, instance_pool(zone_id=zone.id, teardown=False)
        )
        instance_pool_id = instance_pool.id
        instance_pool_zone_id = instance_pool.zone.id

        # Let's wait until the Instance Pool fixture state stabilizes before destroying
        # it, otherwise we'll likely run into a race condition
        t = datetime.now()
        while exo.compute.cs.getInstancePool(
            zoneid=zone.id, id=instance_pool.id, fetch_list=True
        )[0]["state"] == "scaling-up" and datetime.now() - t < timedelta(minutes=5):
            sleep(10)

        instance_pool.delete()
        assert instance_pool.id is None

        with pytest.raises(CloudStackApiException) as excinfo:
            exo.compute.cs.getInstancePool(
                id=instance_pool_id, zoneid=instance_pool_zone_id, fetch_list=True
            )

    def test_properties(self, exo, zone, aag, sg, privnet, instance_pool):
        zone = Zone._from_cs(zone())
        anti_affinity_group = AntiAffinityGroup._from_cs(exo.compute, aag())
        security_group = SecurityGroup._from_cs(exo.compute, sg())
        private_network = PrivateNetwork._from_cs(exo.compute, privnet())
        instance_pool = InstancePool._from_cs(
            exo.compute,
            instance_pool(
                zone_id=zone.id,
                size=1,
                anti_affinity_groups=[anti_affinity_group.id],
                security_groups=[security_group.id],
                private_networks=[private_network.id],
                teardown=False,
            ),
        )

        instance_pool_instances = list(instance_pool.instances)
        assert len(instance_pool_instances) == 1
        assert instance_pool_instances[0].instance_pool.id == instance_pool.id
        assert instance_pool.state in ["scaling-up", "running"]

        instance_pool_anti_affinity_groups = list(instance_pool.anti_affinity_groups)
        assert len(instance_pool_anti_affinity_groups) == 1
        assert instance_pool_anti_affinity_groups[0].id == anti_affinity_group.id

        instance_pool_security_groups = list(instance_pool.security_groups)
        assert len(instance_pool_security_groups) == 1
        assert instance_pool_security_groups[0].id == security_group.id

        instance_pool_private_networks = list(instance_pool.private_networks)
        assert len(instance_pool_private_networks) == 1
        assert instance_pool_private_networks[0].id == private_network.id

        # We have to delete the fixture Instance Pool and ensure its member instance is
        # actually deleted here because of a race condition with the AAG/SG fixtures
        # teardown than will fail because it can't delete itself while still used by an
        # instance ¯\_(ツ)_/¯
        exo.compute.cs.destroyInstancePool(id=instance_pool.id, zoneid=zone.id)
        t = datetime.now()
        try:
            while exo.compute.get_instance(
                zone=zone, id=instance_pool_instances[0].id
            ) != None and datetime.now() - t < timedelta(minutes=5):
                sleep(10)
        except ResourceNotFoundError:
            return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.compute import *


class TestComputeSSHKey:
    def test_delete(self, exo, sshkey):
        ssh_key = SSHKey._from_cs(exo.compute, sshkey(teardown=False))
        ssh_key_name = ssh_key.name

        ssh_key.delete()
        assert ssh_key.name is None

        res = exo.compute.cs.listSSHKeyPairs(name=ssh_key_name, fetch_list=True)
        assert len(res) == 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.runstatus import *
from .conftest import _random_str


class TestRunstatus:
    ### Page

    def test_create_page(self, exo, test_prefix):
        page_name = "-".join([test_prefix, _random_str()])

        page = exo.runstatus.create_page(name=page_name)
        assert page.id > 0
        assert page.name == page_name

        exo.runstatus._delete(url="/pages/{p}".format(p=page_name))

    def test_list_pages(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())

        pages = list(exo.runstatus.list_pages())
        # We cannot guarantee that there will be only our resources in the
        # testing environment, so we ensure we get at least our fixture page
        assert len(pages) >= 1

    def test_get_page(self, exo, runstatus_page):
        page1 = Page._from_rs(exo.runstatus, runstatus_page())

        page = exo.runstatus.get_page(name=page1.name)
        assert page.name == page1.name

        with pytest.raises(ResourceNotFoundError) as excinfo:
            page = exo.runstatus.get_page(name="lolnope")
            assert page is None
        assert excinfo.type == ResourceNotFoundError



# -*- coding: utf-8 -*-

"""
This submodule represents the Exoscale Runstatus API.
"""

import attr
from datetime import datetime
from exoscale_auth import ExoscaleAuth
from . import API, Resource, APIException, RequestError, ResourceNotFoundError

_SUPPORTED_INCIDENT_STATES = {"major_outage", "partial_outage", "degraded_performance"}
_SUPPORTED_INCIDENT_STATUSES = {"investigating", "identified", "monitoring"}
_SUPPORTED_MAINTENANCE_STATUSES = {"scheduled", "in-progress"}


def rstime_to_datetime(rstime):
    """
    Parses a Runstatus API timestamp and returns the corresponding datetime object.
    """

    try:
        return datetime.strptime(rstime, "%Y-%m-%dT%H:%M:%S.%fZ")
    except ValueError:
        return datetime.strptime(rstime, "%Y-%m-%dT%H:%M:%SZ")


@attr.s
class Service(Resource):
    """
    A Runstatus service.

    Attributes:
        id (int): the service unique identifier
        name (str): the service name
        page (Page): the page the incident belongs to
    """

    runstatus = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    name = attr.ib()
    page = attr.ib(repr=False)

    @classmethod
    def _from_rs(cls, runstatus, res, page):
        return cls(
            runstatus, res, id=res["url"].split("/")[-1], name=res["name"], page=page
        )

    @property
    def state(self):
        """
        Service state.

        Returns:
            str: the service state

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        if "state" in self.res:
            return self.res["state"]

        res = self.runstatus._get(
            url="/pages/{p}/services/{s}".format(p=self.page.name, s=self.id)
        )

        return res["state"]

    def delete(self):
        """
        Delete the service.

        Returns:
            None
        """

        self.runstatus._delete(
            url="/pages/{p}/services/{s}".format(p=self.page.name, s=self.id)
        )

        self.runstatus = None
        self.res = None
        self.id = None
        self.name = None
        self.page = None


@attr.s
class IncidentEvent(Resource):
    """
    A Runstatus incident event.

    Attributes:
        date (datetime.datetime): the event date
        description (str): the event description
        status (str): the target status for the incident
        state (str): the target state for the services impacted
        incident (Incident): the incident the event belongs to
    """

    runstatus = attr.ib(repr=False)
    res = attr.ib(repr=False)
    date = attr.ib(repr=False)
    description = attr.ib(repr=False)
    status = attr.ib(repr=False)
    state = attr.ib(repr=False)
    incident = attr.ib(repr=False)


@attr.s
class Incident(Resource):
    """
    A Runstatus incident.

    Attributes:
        id (int): the incident unique identifier
        title (str): a incident title
        start_date (datetime.datetime): the incident start date
        end_date (datetime.datetime): the incident end date
        services ([str]): a list of services impacted by the incident
        state (str): the incident state
        status (str): the incident status
        page (Page): the page the incident belongs to
    """

    runstatus = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    page = attr.ib(repr=False)
    state = attr.ib(repr=False)
    status = attr.ib(repr=False)
    start_date = attr.ib(repr=False)
    end_date = attr.ib(default=None, repr=False)
    title = attr.ib(default=None, repr=False)
    services = attr.ib(default=None, repr=False)

    @classmethod
    def _from_rs(cls, runstatus, res, page):
        return cls(
            runstatus,
            res,
            id=res["id"],
            page=page,
            state=res["state"],
            status=res["status"],
            start_date=rstime_to_datetime(res["start_date"]),
            end_date=rstime_to_datetime(res["end_date"])
            if res["end_date"] is not None
            else None,
            title=res["title"],
            services=res["services"],
        )

    @property
    def events(self):
        """
        Incident event stream.

        Yields:
            IncidentEvent: the next incident event

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        res = self.runstatus._get(
            url="/pages/{p}/incidents/{i}/events".format(p=self.page.name, i=self.id)
        )

        for i in res.json().get("results", []):
            yield IncidentEvent(
                self.runstatus,
                i,
                date=rstime_to_datetime(i["created"]),
                description=i["text"],
                status=i["status"],
                state=i["state"],
                incident=self,
            )

    def add_event(self, description, state=None, status=None):
        """
        Update the incident event stream.

        Parameters:
            description (str): the event description
            state (str): a new incident state to set
            status (str): a new incident status to set

        Returns:
            None
        """

        if state is not None and state not in _SUPPORTED_INCIDENT_STATES:
            raise ValueError(
                "unsupported state; supported states are: {}".format(
                    ",".join(_SUPPORTED_INCIDENT_STATES)
                )
            )

        if status is not None and status not in _SUPPORTED_INCIDENT_STATUSES:
            raise ValueError(
                "unsupported status; supported statuses are: {}".format(
                    ",".join(_SUPPORTED_INCIDENT_STATUSES)
                )
            )

        self.runstatus._post(
            url="/pages/{p}/incidents/{i}/events".format(p=self.page.name, i=self.id),
            json={
                "text": description,
                "status": status if status is not None else self.status,
                "state": state if state is not None else self.state,
            },
        )

        self.state = state if state is not None else self.state
        self.status = status if status is not None else self.status

    def update(self, title=None, services=None):
        """
        Update the incident properties.

        Parameters:
            title (str): an incident title
            services ([str]): a list of services impacted by the incident

        Returns:
            None
        """

        json = {}
        if title is not None:
            json["title"] = title
        if services is not None:
            json["services"] = services

        self.runstatus._patch(
            url="/pages/{p}/incidents/{i}".format(p=self.page.name, i=self.id),
            json=json,
        )

        if title is not None:
            self.title = title
        if services is not None:
            self.services = services

    def close(self, description):
        """
        Close the incident.

        Parameters:
            description (str): the incident closing description

        Returns:
            None
        """

        self.runstatus._post(
            url="/pages/{p}/incidents/{i}/events".format(p=self.page.name, i=self.id),
            json={"text": description, "status": "resolved", "state": "operational"},
        )


@attr.s
class MaintenanceEvent(Resource):
    """
    A Runstatus maintenance event.

    Attributes:
        date (datetime.datetime): the event date
        description (str): the event description
        status (str): the target status for the maintenance
        maintenance (Maintenance): the maintenance the event belongs to
    """

    runstatus = attr.ib(repr=False)
    res = attr.ib(repr=False)
    date = attr.ib(repr=False)
    description = attr.ib(repr=False)
    status = attr.ib(repr=False)
    maintenance = attr.ib(repr=False)


@attr.s
class Maintenance(Resource):
    """
    A Runstatus maintenance.

    Attributes:
        id (int): the maintenance unique identifier
        title (str): a maintenance title
        description (str): a maintenance description
        services ([str]): a list of services impacted by the maintenance
        start_date (datetime.datetime): a maintenance start date
        end_date (datetime.datetime): a maintenance end date
        status (str): the maintenance status
        page (Page): the page the maintenance belongs to
    """

    runstatus = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    page = attr.ib(repr=False)
    status = attr.ib(repr=False)
    start_date = attr.ib(repr=False)
    end_date = attr.ib(repr=False)
    title = attr.ib(default=None, repr=False)
    description = attr.ib(default=None, repr=False)
    services = attr.ib(default=None, repr=False)

    @classmethod
    def _from_rs(cls, runstatus, res, page):
        return cls(
            runstatus,
            res,
            id=res["id"],
            page=page,
            status=res["status"],
            start_date=rstime_to_datetime(res["start_date"]),
            end_date=rstime_to_datetime(res["end_date"]),
            title=res["title"],
            description=res["description"],
            services=res["services"],
        )

    @property
    def events(self):
        """
        Maintenance event stream.

        Yields:
            MaintenanceEvent: the next maintenance event

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        res = self.runstatus._get(
            url="/pages/{p}/maintenances/{m}/events".format(p=self.page.name, m=self.id)
        )

        for i in res.json().get("results", []):
            yield MaintenanceEvent(
                self.runstatus,
                i,
                date=rstime_to_datetime(i["created"]),
                description=i["text"],
                status=i["status"],
                maintenance=self,
            )

    def add_event(self, description, status=None):
        """
        Update the maintenance event stream.

        Parameters:
            description (str): the event description
            status (str): a new maintenance status to set

        Returns:
            None
        """

        if status is not None and status not in _SUPPORTED_MAINTENANCE_STATUSES:
            raise ValueError(
                "unsupported status; supported statuses are: {}".format(
                    ",".join(_SUPPORTED_MAINTENANCE_STATUSES)
                )
            )

        self.runstatus._post(
            url="/pages/{p}/maintenances/{m}/events".format(
                p=self.page.name, m=self.id
            ),
            json={
                "text": description,
                "status": status if status is not None else self.status,
            },
        )

        self.status = status if status is not None else self.status

    def update(
        self,
        title=None,
        description=None,
        start_date=None,
        end_date=None,
        services=None,
    ):
        """
        Update the maintenance properties.

        Parameters:
            title (str): a maintenance title
            description (str): a maintenance description
            services ([str]): a list of services impacted by the maintenance
            start_date (datetime.datetime): a maintenance start date
            end_date (datetime.datetime): a maintenance end date

        Returns:
            None
        """

        json = {}
        if title is not None:
            json["title"] = title
        if description is not None:
            json["description"] = description
        if services is not None:
            json["services"] = services
        if start_date is not None:
            json["start_date"] = start_date.isoformat()
        if end_date is not None:
            json["end_date"] = end_date.isoformat()

        self.runstatus._patch(
            url="/pages/{p}/maintenances/{i}".format(p=self.page.name, i=self.id),
            json=json,
        )

        if title is not None:
            self.title = title
        if description is not None:
            self.description = description
        if services is not None:
            self.services = services
        if start_date is not None:
            self.start_date = start_date
        if end_date is not None:
            self.end_date = end_date

    def close(self, description):
        """
        Close the maintenance.

        Parameters:
            description (str): the maintenance closing description

        Returns:
            None
        """

        self.runstatus._post(
            url="/pages/{p}/maintenances/{m}/events".format(
                p=self.page.name, m=self.id
            ),
            json={"text": description, "status": "completed"},
        )


@attr.s
class Page(Resource):
    """
    A Runstatus page.

    Attributes:
        id (int): the page unique identifier
        name (str): the page name
        title (str): a page title
        default_status_message (str): a default "OK" status message
        custom_domain (str): a custom page domain name
        time_zone (str): a time zone

    Note:
        The expected time zone format is the tz database (a.k.a "tzdata" or "Olson"
        database): https://en.wikipedia.org/wiki/Tz_database
    """

    runstatus = attr.ib(repr=False)
    res = attr.ib(repr=False)
    id = attr.ib()
    name = attr.ib()
    title = attr.ib(default=None, repr=False)
    default_status_message = attr.ib(default=None, repr=False)
    custom_domain = attr.ib(default=None, repr=False)
    time_zone = attr.ib(default=None, repr=False)

    @classmethod
    def _from_rs(cls, runstatus, res):
        return cls(runstatus, res, id=res["id"], name=res["subdomain"])

    @property
    def services(self):
        """
        Page services.

        Yields:
            Service: the next service

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        res = self.runstatus._get(url="/pages/{p}/services".format(p=self.name))

        for i in res.json().get("results", []):
            yield Service._from_rs(self.runstatus, i, page=self)

    def add_service(self, name):
        """
        Add a service to the page.

        Parameters:
            name (str): the name of the service

        Returns:
            None
        """

        self.runstatus._post(
            url="/pages/{p}/services".format(p=self.name), json={"name": name}
        )

    @property
    def incidents(self):
        """
        Page incidents.

        Yields:
            Incident: the next incident

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        res = self.runstatus._get(url="/pages/{p}/incidents".format(p=self.name))

        for i in res.json().get("results", []):
            yield Incident._from_rs(self.runstatus, i, self)

    def add_incident(self, title, description, state, status, services=None):
        """
        Open a new incident.

        Parameters:
            title (str): the incident title
            description (str): the incident description
            state (str): the incident state
            status (str): the incident status
            services ([str]): the names of services impacted by the incident

        Returns:
            None
        """

        if services is None:
            services = []

        if state not in _SUPPORTED_INCIDENT_STATES:
            raise ValueError(
                "unsupported state; supported states are: {}".format(
                    ",".join(_SUPPORTED_INCIDENT_STATES)
                )
            )

        if status not in _SUPPORTED_INCIDENT_STATUSES:
            raise ValueError(
                "unsupported status; supported statuses are: {}".format(
                    ",".join(_SUPPORTED_INCIDENT_STATUSES)
                )
            )

        self.runstatus._post(
            url="/pages/{p}/incidents".format(p=self.name),
            json={
                "title": title,
                "status_text": description,
                "status": status,
                "state": state,
                "services": services,
            },
        )

    @property
    def maintenances(self):
        """
        Page maintenance.

        Yields:
            Maintenance: the next maintenance

        Note:
            This property value is dynamically retrieved from the API, incurring extra
            latency.
        """

        res = self.runstatus._get(url="/pages/{p}/maintenances".format(p=self.name))

        for i in res.json().get("results", []):
            yield Maintenance._from_rs(self.runstatus, i, self)

    def add_maintenance(self, title, description, start_date, end_date, services=None):
        """
        Open a new maintenance.

        Parameters:
            title (str): the maintenance title
            description (str): the maintenance description
            start_date (datetime.datetime): the maintenance start date
            end_date (datetime.datetime): the maintenance end date
            services ([str]): the names of services impacted by the maintenance

        Returns:
            None
        """

        if services is None:
            services = []

        self.runstatus._post(
            url="/pages/{p}/maintenances".format(p=self.name),
            json={
                "title": title,
                "description": description,
                "start_date": start_date.isoformat(),
                "end_date": end_date.isoformat(),
                "services": services,
            },
        )

    def update(
        self,
        title=None,
        default_status_message=None,
        custom_domain=None,
        time_zone=None,
    ):
        """
        Update the page properties.

        Parameters:
            title (str): a page title
            default_status_message (str): a default "OK" status message
            custom_domain (str): a custom page domain name
            time_zone (str): a time zone

        Returns:
            None

        Note:
            The expected time zone format is the tz database (a.k.a "tzdata" or "Olson"
            database): https://en.wikipedia.org/wiki/Tz_database
        """

        json = {}
        if title is not None:
            json["title"] = title
        if default_status_message is not None:
            json["ok_text"] = default_status_message
        if custom_domain is not None:
            json["domain"] = custom_domain
        if time_zone is not None:
            json["time_zone"] = time_zone

        self.runstatus._patch(url="/pages/{p}".format(p=self.name), json=json)

        if title is not None:
            self.title = title
        if default_status_message is not None:
            self.default_status_message = default_status_message
        if custom_domain is not None:
            self.custom_domain = custom_domain
        if time_zone is not None:
            self.time_zone = time_zone

    def delete(self):
        """
        Delete the page.

        Returns:
            None
        """

        self.runstatus._delete(url="/pages/{p}".format(p=self.name))

        self.runstatus = None
        self.res = None
        self.id = None
        self.name = None


class RunstatusAPI(API):
    """
    An Exoscale Runstatus API client.

    Parameters:
        key (str): the Runstatus API key
        secret (str): the Runstatus API secret
        endpoint (str): the Runstatus API endpoint
        max_retries (int): the API HTTP session retry policy number of retries to allow
        trace (bool): API request/response tracing flag
    """

    def __init__(
        self,
        key,
        secret,
        endpoint="https://api.runstatus.com",
        max_retries=None,
        trace=False,
    ):
        super().__init__(
            endpoint=endpoint,
            key=key,
            secret=secret,
            max_retries=max_retries,
            trace=trace,
        )

        self.auth = ExoscaleAuth(self.key, self.secret)

    def __repr__(self):
        return "RunstatusAPI(endpoint='{}' key='{}')".format(self.endpoint, self.key)

    def __str__(self):
        return self.__repr__()

    def _check_api_response(self, res, *args, **kwargs):
        """
        Check the API response and raise an exception depending on the status code.
        """

        if res.status_code >= 500:
            raise APIException(res.text)

        if res.status_code == 404:
            raise ResourceNotFoundError

        if res.status_code >= 400:
            raise RequestError(str(res.text))

    def _send(self, **kwargs):
        """
        Send a free-form HTTP request.

        Parameters:
            kwargs: request.Request parameters

        Returns:
            request.Response: the HTTP response
        """

        return API.send(
            self, auth=self.auth, hooks={"response": self._check_api_response}, **kwargs
        )

    def _get(self, url, **kwargs):
        """
        Send an HTTP GET request to the Runstatus API.

        Parameters:
            url (str): the URL to request
            kwargs: request.Request parameters

        Returns:
            request.Response: the HTTP response received from the Runstatus API
        """

        return self._send(
            method="GET", url=self.endpoint + "/" + url.lstrip("/"), **kwargs
        )

    def _post(self, url, json, **kwargs):
        """
        Send an HTTP POST request to the Runstatus API.

        Parameters:
            url (str): the URL to request
            json (dict): the dict of parameters to send in JSON-formatted payload
            kwargs: request.Request parameters

        Returns:
            request.Response: the HTTP response received from the Runstatus API
        """

        return self._send(
            method="POST",
            url=self.endpoint + "/" + url.lstrip("/"),
            json=json,
            **kwargs
        )

    def _patch(self, url, json, **kwargs):
        """
        Send an HTTP PATCH request to the Runstatus API.

        Parameters:
            url (str): the URL to request
            json (dict): the dict of parameters to send in JSON-formatted payload
            kwargs: request.Request parameters

        Returns:
            request.Response: the HTTP response received from the Runstatus API
        """

        return self._send(
            method="PATCH",
            url=self.endpoint + "/" + url.lstrip("/"),
            json=json,
            **kwargs
        )

    def _delete(self, url, **kwargs):
        """
        Send an HTTP DELETE request to the Runstatus API.

        Parameters:
            url (str): the URL to request
            kwargs: request.Request parameters

        Returns:
            request.Response: the HTTP response received from the Runstatus API
        """

        return self._send(
            method="DELETE", url=self.endpoint + "/" + url.lstrip("/"), **kwargs
        )

    ## Page

    def create_page(self, name):
        """
        Create a Runstatus page.

        Parameters:
            name (str): the page name

        Returns:
            Page: the Runstatus page created
        """

        res = self._post(url="/pages", json={"name": name, "subdomain": name})

        return Page._from_rs(self, res.json())

    def list_pages(self):
        """
        List Runstatus pages.

        Yields:
            Page: the next Runstatus page
        """

        res = self._get(url="/pages")

        for i in res.json().get("results", []):
            yield Page._from_rs(self, i)

    def get_page(self, name=None, id=None):
        """
        Get a Runstatus page.

        Parameters:
            name (str): a Runstatus page name
            id (int): a Runstatus page unique identifier

        Returns:
            Page: a Runstatus page
        """

        if id is None and name is None:
            raise ValueError("either id or name must be specifed")

        pages = self.list_pages()

        for page in pages:
            if (name and page.name == name) or (id and page.id == id):
                return page

        raise ResourceNotFoundError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.compute import *


class TestComputeSecurityGroupRule:
    def test_delete(self, exo, sg):
        security_group = SecurityGroup._from_cs(exo.compute, sg())

        rule1 = exo.compute.cs.authorizeSecurityGroupIngress(
            securitygroupid=security_group.id,
            cidrlist="0.0.0.0/0",
            protocol="TCP",
            startport=80,
            endport=80,
        )["securitygroup"]["ingressrule"][0]
        rule2 = exo.compute.cs.authorizeSecurityGroupEgress(
            securitygroupid=security_group.id,
            protocol="TCP",
            cidrlist="0.0.0.0/0",
            startport=80,
            endport=80,
        )["securitygroup"]["egressrule"][0]

        res = exo.compute.cs.listSecurityGroups(id=security_group.id, fetch_list=True)
        SecurityGroupRule._from_cs(
            compute=exo.compute, type="ingress", res=res[0]["ingressrule"][0]
        ).delete()
        SecurityGroupRule._from_cs(
            compute=exo.compute, type="egress", res=res[0]["egressrule"][0]
        ).delete()

        res = exo.compute.cs.listSecurityGroups(id=security_group.id, fetch_list=True)
        assert len(res[0]["ingressrule"]) == 0
        assert len(res[0]["egressrule"]) == 0

    def test_parse_port(self):
        rule = SecurityGroupRule(type="ingress")
        assert rule._parse_port() == (None, None)

        rule = SecurityGroupRule(type="ingress", port="80")
        assert rule._parse_port() == (80, 80)

        rule = SecurityGroupRule(type="ingress", port="80-81")
        assert rule._parse_port() == (80, 81)



from setuptools import setup, find_packages

INSTALL_REQUIRES = []
with open("requirements.txt", "r", encoding="utf-8") as f:
    INSTALL_REQUIRES = list(i.rstrip() for i in f.readlines())

EXTRA_REQUIRE = []
with open("requirements.dev.txt", "r", encoding="utf-8") as f:
    EXTRA_REQUIRE = list(i.rstrip() for i in f.readlines())

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="exoscale",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    version="0.5.0",
    license="ISC",
    url="https://github.com/exoscale/python-exoscale",
    author="Exoscale",
    author_email="contact@exoscale.com",
    platforms="any",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires=">=3.6",
    install_requires=INSTALL_REQUIRES,
    extra_require=EXTRA_REQUIRE,
    tests_require=["pytest>=5.0.0"],
    include_package_data=True,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.runstatus import *
from datetime import timezone
from .conftest import _random_str


class TestRunstatusPage:
    def test_add_service(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())

        page.add_service(name="test")

        res = exo.runstatus._get(url="/pages/{p}/services".format(p=page.name)).json()
        assert len(res["results"]) == 1
        assert res["results"][0]["name"] == "test"

    def test_add_incident(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())
        test_incident_title = "Everything's on fire"
        test_incident_description = "It's fine ¯\\_(ツ)_/¯"
        test_incident_state = "major_outage"
        test_incident_status = "identified"
        test_incident_services = ["a", "b", "c"]

        for i in test_incident_services:
            exo.runstatus._post(
                url="/pages/{p}/services".format(p=page.name), json={"name": i}
            )

        page.add_incident(
            title=test_incident_title,
            description=test_incident_description,
            state=test_incident_state,
            status=test_incident_status,
            services=test_incident_services,
        )

        res = exo.runstatus._get(url="/pages/{p}/incidents".format(p=page.name)).json()
        assert len(res["results"]) == 1
        assert res["results"][0]["title"] == test_incident_title
        assert res["results"][0]["state"] == test_incident_state
        assert res["results"][0]["status"] == test_incident_status
        assert res["results"][0]["services"] == test_incident_services

    def test_add_maintenance(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())
        test_maintenance_title = "Database server upgrade"
        test_maintenance_description = (
            "We're upgrading the database server hardware to add more memory"
        )
        test_maintenance_service = "db"
        test_maintenance_start_date = datetime(2019, 8, 2, 4, 0)
        test_maintenance_end_date = datetime(2019, 8, 2, 5, 0)

        exo.runstatus._post(
            url="/pages/{p}/services".format(p=page.name),
            json={"name": test_maintenance_service},
        )

        page.add_maintenance(
            title=test_maintenance_title,
            description=test_maintenance_description,
            start_date=test_maintenance_start_date,
            end_date=test_maintenance_end_date,
            services=[test_maintenance_service],
        )

        res = exo.runstatus._get(
            url="/pages/{p}/maintenances".format(p=page.name)
        ).json()
        assert len(res["results"]) == 1
        assert res["results"][0]["title"] == test_maintenance_title
        assert res["results"][0]["description"] == test_maintenance_description
        assert (
            rstime_to_datetime(res["results"][0]["start_date"])
            == test_maintenance_start_date
        )
        assert (
            rstime_to_datetime(res["results"][0]["end_date"])
            == test_maintenance_end_date
        )
        assert res["results"][0]["services"] == [test_maintenance_service]

    def test_delete(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page(teardown=False))

        page.delete()
        assert page.id is None

        with pytest.raises(ResourceNotFoundError) as excinfo:
            res = exo.runstatus._get(url="/pages/{p}".format(p=page.name))
        assert excinfo.type == ResourceNotFoundError

    def test_update(self, exo, runstatus_page, test_prefix):
        page = Page._from_rs(exo.runstatus, runstatus_page())
        test_page_title_edited = "python-exoscale"
        test_page_default_status_message_edited = "It's all good in the hood"
        test_page_custom_domain_edited = (
            "-".join([test_prefix, _random_str()]) + ".example.net"
        )
        test_page_time_zone_edited = "Europe/Zurich"

        page.update(
            title=test_page_title_edited,
            default_status_message=test_page_default_status_message_edited,
            custom_domain=test_page_custom_domain_edited,
            time_zone=test_page_time_zone_edited,
        )

        res = exo.runstatus._get(url="/pages/{p}".format(p=page.name)).json()
        assert res["title"] == test_page_title_edited
        assert page.title == test_page_title_edited
        assert res["ok_text"] == test_page_default_status_message_edited
        assert page.default_status_message == test_page_default_status_message_edited
        assert res["domain"] == test_page_custom_domain_edited
        assert page.custom_domain == test_page_custom_domain_edited
        assert res["time_zone"] == test_page_time_zone_edited
        assert page.time_zone == test_page_time_zone_edited

    def test_properties(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())
        test_page_service_name = "python-exoscale"
        test_incident_title = "Everything's on fire"
        test_incident_description = "It's fine ¯\\_(ツ)_/¯"
        test_incident_state = "major_outage"
        test_incident_status = "identified"
        test_maintenance_title = "Database server upgrade"
        test_maintenance_description = (
            "We're upgrading the database server hardware to add more memory"
        )
        test_maintenance_start_date = datetime(2019, 8, 2, 4, 0)
        test_maintenance_end_date = datetime(2019, 8, 2, 5, 0)

        exo.runstatus._post(
            url="/pages/{p}/services".format(p=page.name),
            json={"name": test_page_service_name},
        )
        page_services = list(page.services)
        assert len(page_services)
        assert page_services[0].name == test_page_service_name

        exo.runstatus._post(
            url="/pages/{p}/incidents".format(p=page.name),
            json={
                "title": test_incident_title,
                "status_text": test_incident_description,
                "status": test_incident_status,
                "state": test_incident_state,
                "services": [test_page_service_name],
            },
        )
        incidents = list(page.incidents)
        assert len(incidents) == 1
        assert incidents[0].id > 0
        assert incidents[0].title == test_incident_title
        assert type(incidents[0].start_date) == datetime
        assert incidents[0].state == test_incident_state
        assert incidents[0].status == test_incident_status
        assert incidents[0].services == [test_page_service_name]
        assert incidents[0].page.id == page.id

        exo.runstatus._post(
            url="/pages/{p}/maintenances".format(p=page.name),
            json={
                "title": test_maintenance_title,
                "description": test_maintenance_description,
                "start_date": test_maintenance_start_date.isoformat(),
                "end_date": test_maintenance_end_date.isoformat(),
                "services": [test_page_service_name],
            },
        )
        maintenances = list(page.maintenances)
        assert len(maintenances) == 1
        assert maintenances[0].id > 0
        assert maintenances[0].title == test_maintenance_title
        assert maintenances[0].start_date == test_maintenance_start_date
        assert maintenances[0].end_date == test_maintenance_end_date
        assert maintenances[0].services == [test_page_service_name]
        assert maintenances[0].page.id == page.id



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.runstatus import *
from datetime import timezone


class TestRunstatusPageMaintenance:
    def test_add_event(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())
        test_maintenance_title = "Database server upgrade"
        test_maintenance_description = (
            "We're upgrading the database server hardware to add more memory"
        )
        test_maintenance_service = "db"
        test_maintenance_start_date = datetime(2019, 8, 2, 4, 0)
        test_maintenance_end_date = datetime(2019, 8, 2, 5, 0)

        exo.runstatus._post(
            url="/pages/{p}/services".format(p=page.name),
            json={"name": test_maintenance_service},
        )

        exo.runstatus._post(
            url="/pages/{p}/maintenances".format(p=page.name),
            json={
                "title": test_maintenance_title,
                "description": test_maintenance_description,
                "start_date": test_maintenance_start_date.isoformat(),
                "end_date": test_maintenance_end_date.isoformat(),
                "services": [test_maintenance_service],
            },
        )

        res = exo.runstatus._get(
            url="/pages/{p}/maintenances".format(p=page.name)
        ).json()
        maintenance = Maintenance._from_rs(exo.runstatus, res["results"][0], page)

        maintenance.add_event(description="Stopping server", status="in-progress")
        maintenance.add_event(description="Upgrading memory")
        maintenance.add_event(description="Restarting server")

        res = exo.runstatus._get(
            url="/pages/{p}/maintenances/{i}/events".format(
                p=page.name, i=maintenance.id
            )
        ).json()
        assert len(res["results"]) == 3
        assert res["results"][0]["text"] == "Restarting server"
        assert res["results"][0]["status"] == "in-progress"
        assert res["results"][1]["text"] == "Upgrading memory"
        assert res["results"][1]["status"] == "in-progress"
        assert res["results"][2]["text"] == "Stopping server"
        assert res["results"][2]["status"] == "in-progress"

    def test_update(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())
        test_maintenance_title = "Database server upgrade"
        test_maintenance_title_edited = "Database server upgrade (edited)"
        test_maintenance_description = (
            "We're upgrading the database server hardware to add more memory"
        )
        test_maintenance_description_edited = (
            "We're upgrading the database server hardware to add more memory (edited)"
        )
        test_maintenance_service = "db1"
        test_maintenance_service_edited = "db2"
        test_maintenance_start_date = datetime(2019, 8, 2, 4, 0)
        test_maintenance_start_date_edited = datetime(2019, 8, 2, 14, 0)
        test_maintenance_end_date = datetime(2019, 8, 2, 5, 0)
        test_maintenance_end_date_edited = datetime(2019, 8, 2, 15, 0)

        for i in [test_maintenance_service, test_maintenance_service_edited]:
            exo.runstatus._post(
                url="/pages/{p}/services".format(p=page.name), json={"name": i}
            )

        exo.runstatus._post(
            url="/pages/{p}/maintenances".format(p=page.name),
            json={
                "title": test_maintenance_title,
                "description": test_maintenance_description,
                "start_date": test_maintenance_start_date.isoformat(),
                "end_date": test_maintenance_end_date.isoformat(),
                "services": [test_maintenance_service],
            },
        )

        res = exo.runstatus._get(
            url="/pages/{p}/maintenances".format(p=page.name)
        ).json()
        maintenance = Maintenance._from_rs(exo.runstatus, res["results"][0], page)

        maintenance.update(
            title=test_maintenance_title_edited,
            description=test_maintenance_description_edited,
            start_date=test_maintenance_start_date_edited,
            end_date=test_maintenance_end_date_edited,
            services=[test_maintenance_service_edited],
        )

        res = exo.runstatus._get(
            url="/pages/{p}/maintenances/{m}".format(p=page.name, m=maintenance.id)
        ).json()
        assert res["title"] == test_maintenance_title_edited
        assert maintenance.title == test_maintenance_title_edited
        assert res["description"] == test_maintenance_description_edited
        assert maintenance.description == test_maintenance_description_edited
        assert (
            rstime_to_datetime(res["start_date"]) == test_maintenance_start_date_edited
        )
        assert maintenance.start_date == test_maintenance_start_date_edited
        assert rstime_to_datetime(res["end_date"]) == test_maintenance_end_date_edited
        assert maintenance.end_date == test_maintenance_end_date_edited
        assert res["services"] == [test_maintenance_service_edited]
        assert maintenance.services == [test_maintenance_service_edited]

    def test_close(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())
        test_maintenance_title = "Database server upgrade"
        test_maintenance_description = (
            "We're upgrading the database server hardware to add more memory"
        )
        test_maintenance_service = "db"
        test_maintenance_start_date = datetime(2019, 8, 2, 4, 0)
        test_maintenance_end_date = datetime(2019, 8, 2, 5, 0)

        exo.runstatus._post(
            url="/pages/{p}/services".format(p=page.name),
            json={"name": test_maintenance_service},
        )

        exo.runstatus._post(
            url="/pages/{p}/maintenances".format(p=page.name),
            json={
                "title": test_maintenance_title,
                "description": test_maintenance_description,
                "start_date": test_maintenance_start_date.isoformat(),
                "end_date": test_maintenance_end_date.isoformat(),
                "services": [test_maintenance_service],
            },
        )

        res = exo.runstatus._get(
            url="/pages/{p}/maintenances".format(p=page.name)
        ).json()
        maintenance = Maintenance._from_rs(exo.runstatus, res["results"][0], page)

        maintenance.close("We're done here")

        res = exo.runstatus._get(
            url="/pages/{p}/maintenances/{i}/events".format(
                p=page.name, i=maintenance.id
            )
        ).json()
        assert len(res["results"]) == 1
        assert res["results"][0]["text"] == "We're done here"
        assert res["results"][0]["status"] == "completed"

    def test_properties(self, exo, runstatus_page):
        page = Page._from_rs(exo.runstatus, runstatus_page())
        test_maintenance_title = "Database server upgrade"
        test_maintenance_description = (
            "We're upgrading the database server hardware to add more memory"
        )
        test_maintenance_services = ["db1", "db2"]
        test_maintenance_start_date = datetime(2019, 8, 2, 4, 0)
        test_maintenance_end_date = datetime(2019, 8, 2, 5, 0)

        for i in test_maintenance_services:
            exo.runstatus._post(
                url="/pages/{p}/services".format(p=page.name), json={"name": i}
            )

        exo.runstatus._post(
            url="/pages/{p}/maintenances".format(p=page.name),
            json={
                "title": test_maintenance_title,
                "description": test_maintenance_description,
                "start_date": test_maintenance_start_date.isoformat(),
                "end_date": test_maintenance_end_date.isoformat(),
                "services": [test_maintenance_services[0]],
            },
        )

        res = exo.runstatus._get(
            url="/pages/{p}/maintenances".format(p=page.name)
        ).json()
        maintenance = Maintenance._from_rs(exo.runstatus, res["results"][0], page)

        for i in [1, 2, 3]:
            exo.runstatus._post(
                url="/pages/{p}/maintenances/{i}/events".format(
                    p=page.name, i=maintenance.id
                ),
                json={"text": "Update #{}".format(i), "status": maintenance.status},
            )

        maintenance_events = list(maintenance.events)
        assert len(maintenance_events) == 3
        assert maintenance_events[0].description == "Update #3"
        assert maintenance_events[1].description == "Update #2"
        assert maintenance_events[2].description == "Update #1"



"""
From https://github.com/justiniso/polling

The MIT License (MIT)

Copyright (c) 2015 Justin Iso

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

"""Polling module containing all exceptions and helpers used for the polling function"""

__version__ = '0.3.1'

import time
try:
    from Queue import Queue
except ImportError:
    from queue import Queue


class PollingException(Exception):
    """Base exception that stores all return values of attempted polls"""
    def __init__(self, values, last=None):
        self.values = values
        self.last = last


class TimeoutException(PollingException):
    """Exception raised if polling function times out"""


class MaxCallException(PollingException):
    """Exception raised if maximum number of iterations is exceeded"""


def step_constant(step):
    """Use this function when you want the step to remain fixed in every iteration (typically good for
    instances when you know approximately how long the function should poll for)"""
    return step


def step_linear_double(step):
    """Use this function when you want the step to double each iteration (e.g. like the way ArrayList works in
    Java). Note that this can result in very long poll times after a few iterations"""
    return step * 2


def is_truthy(val):
    """Use this function to test if a return value is truthy"""
    return bool(val)


def poll(target, step, args=(), kwargs=None, timeout=None, max_tries=None, check_success=is_truthy,
         step_function=step_constant, ignore_exceptions=(), poll_forever=False, collect_values=None, *a, **k):
    """Poll by calling a target function until a certain condition is met. You must specify at least a target
    function to be called and the step -- base wait time between each function call.

    :param step: Step defines the amount of time to wait (in seconds)
    :param args: Arguments to be passed to the target function
    :type kwargs: dict
    :param kwargs: Keyword arguments to be passed to the target function
    :param timeout: The target function will be called until the time elapsed is greater than the maximum timeout
    (in seconds). NOTE that the actual execution time of the function *can* exceed the time specified in the timeout.
    For instance, if the target function takes 10 seconds to execute and the timeout is 21 seconds, the polling
    function will take a total of 30 seconds (two iterations of the target --20s which is less than the timeout--21s,
    and a final iteration)
    :param max_tries: Maximum number of times the target function will be called before failing
    :param check_success: A callback function that accepts the return value of the target function. It should
    return true if you want the polling function to stop and return this value. It should return false if you want it
    to continue executing. The default is a callback that tests for truthiness (anything not False, 0, or empty
    collection).
    :param step_function: A callback function that accepts each iteration's "step." By default, this is constant,
    but you can also pass a function that will increase or decrease the step. As an example, you can increase the wait
    time between calling the target function by 10 seconds every iteration until the step is 100 seconds--at which
    point it should remain constant at 100 seconds

    >>> def my_step_function(step):
    >>>     step += 10
    >>>     return max(step, 100)

    :type ignore_exceptions: tuple
    :param ignore_exceptions: You can specify a tuple of exceptions that should be caught and ignored on every
    iteration. If the target function raises one of these exceptions, it will be caught and the exception
    instance will be pushed to the queue of values collected during polling. Any other exceptions raised will be
    raised as normal.
    :param poll_forever: If set to true, this function will retry until an exception is raised or the target's
    return value satisfies the check_success function. If this is not set, then a timeout or a max_tries must be set.
    :type collect_values: Queue
    :param collect_values: By default, polling will create a new Queue to store all of the target's return values.
    Optionally, you can specify your own queue to collect these values for access to it outside of function scope.
    :return: Polling will return first value from the target function that meets the condions of the check_success
    callback. By default, this will be the first value that is not None, 0, False, '', or an empty collection.
    """

    assert (timeout is not None or max_tries is not None) or poll_forever, \
        ('You did not specify a maximum number of tries or a timeout. Without either of these set, the polling '
         'function will poll forever. If this is the behavior you want, pass "poll_forever=True"')

    assert not ((timeout is not None or max_tries is not None) and poll_forever), \
        'You cannot specify both the option to poll_forever and max_tries/timeout.'

    kwargs = kwargs or dict()
    values = collect_values or Queue()

    max_time = time.time() + timeout if timeout else None
    tries = 0

    last_item = None
    while True:

        if max_tries is not None and tries >= max_tries:
            raise MaxCallException(values, last_item)

        try:
            val = target(*args, **kwargs)
            last_item = val
        except ignore_exceptions as e:
            last_item = e
        else:
            # Condition passes, this is the only "successful" exit from the polling function
            if check_success(val):
                return val

        values.put(last_item)
        tries += 1

        # Check the time after to make sure the poll function is called at least once
        if max_time is not None and time.time() >= max_time:
            raise TimeoutException(values, last_item)

        time.sleep(step)
        step = step_function(step)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api.compute import *


class TestComputeElasticIP:
    def test_attach_instance(self, exo, eip, instance):
        elastic_ip = ElasticIP._from_cs(exo.compute, eip())
        instance = Instance._from_cs(exo.compute, instance())

        elastic_ip.attach_instance(instance)

        res = exo.compute.cs.listNics(virtualmachineid=instance.id, fetch_list=True)
        assert (
            instance._default_nic(res)["secondaryip"][0]["ipaddress"]
            == elastic_ip.address
        )

    def test_detach_instance(self, exo, eip, instance):
        elastic_ip = ElasticIP._from_cs(exo.compute, eip())
        instance = Instance._from_cs(exo.compute, instance())

        res = exo.compute.cs.listNics(virtualmachineid=instance.id, fetch_list=True)
        exo.compute.cs.addIpToNic(
            nicid=instance._default_nic(res)["id"], ipaddress=elastic_ip.address
        )

        res = exo.compute.cs.listNics(virtualmachineid=instance.id, fetch_list=True)
        assert (
            instance._default_nic(res)["secondaryip"][0]["ipaddress"]
            == elastic_ip.address
        )

        elastic_ip.detach_instance(instance)

        res = exo.compute.cs.listNics(virtualmachineid=instance.id, fetch_list=True)
        assert "secondaryip" not in instance._default_nic(res)

    def test_set_reverse_dns(self, exo, eip, test_reverse_dns):
        elastic_ip = ElasticIP._from_cs(exo.compute, eip())

        elastic_ip.set_reverse_dns(record=test_reverse_dns)

        res = exo.compute.cs.queryReverseDnsForPublicIpAddress(id=elastic_ip.id)
        assert res["publicipaddress"]["reversedns"][0]["domainname"] == test_reverse_dns

    def test_unset_reverse_dns(self, exo, eip, test_reverse_dns):
        elastic_ip = ElasticIP._from_cs(exo.compute, eip())

        res = exo.compute.cs.updateReverseDnsForPublicIpAddress(
            id=elastic_ip.id, domainname=test_reverse_dns
        )
        assert res["publicipaddress"]["reversedns"][0]["domainname"] == test_reverse_dns

        elastic_ip.unset_reverse_dns()

        res = exo.compute.cs.queryReverseDnsForPublicIpAddress(id=elastic_ip.id)
        assert len(res["publicipaddress"]["reversedns"]) == 0

    def test_description_is_optional(self, exo, eip):
        response = eip()
        # The eip() fixture has a description set - del'ing it here is much cheaper
        # than writing a new fixture just for this particular case.
        del response["description"]
        elastic_ip = ElasticIP._from_cs(exo.compute, response)
        assert elastic_ip.description == ""

    def test_update(self, exo, eip, test_description):
        elastic_ip = ElasticIP._from_cs(exo.compute, eip())
        description_edited = test_description + " (edited)"
        healthcheck_mode = "https"
        healthcheck_port = 443
        healthcheck_path = "/test"
        healthcheck_interval = 5
        healthcheck_timeout = 3
        healthcheck_strikes_ok = 2
        healthcheck_strikes_fail = 1
        healthcheck_tls_sni = "example.net"
        healthcheck_tls_skip_verify = True

        elastic_ip.update(
            description=description_edited,
            healthcheck_mode=healthcheck_mode,
            healthcheck_path=healthcheck_path,
            healthcheck_port=healthcheck_port,
            healthcheck_interval=healthcheck_interval,
            healthcheck_timeout=healthcheck_timeout,
            healthcheck_strikes_ok=healthcheck_strikes_ok,
            healthcheck_strikes_fail=healthcheck_strikes_fail,
            healthcheck_tls_sni=healthcheck_tls_sni,
            healthcheck_tls_skip_verify=healthcheck_tls_skip_verify,
        )

        [res] = exo.compute.cs.listPublicIpAddresses(id=elastic_ip.id, fetch_list=True)
        assert res["description"] == description_edited
        assert elastic_ip.description == description_edited
        assert "healthcheck" in res.keys()
        assert res["healthcheck"]["mode"] == healthcheck_mode
        assert elastic_ip.healthcheck_mode == healthcheck_mode
        assert res["healthcheck"]["path"] == healthcheck_path
        assert elastic_ip.healthcheck_path == healthcheck_path
        assert res["healthcheck"]["port"] == healthcheck_port
        assert elastic_ip.healthcheck_port == healthcheck_port
        assert res["healthcheck"]["interval"] == healthcheck_interval
        assert elastic_ip.healthcheck_interval == healthcheck_interval
        assert res["healthcheck"]["timeout"] == healthcheck_timeout
        assert elastic_ip.healthcheck_timeout == healthcheck_timeout
        assert res["healthcheck"]["strikes-ok"] == healthcheck_strikes_ok
        assert elastic_ip.healthcheck_strikes_ok == healthcheck_strikes_ok
        assert res["healthcheck"]["strikes-fail"] == healthcheck_strikes_fail
        assert elastic_ip.healthcheck_strikes_fail == healthcheck_strikes_fail
        assert res["healthcheck"]["tls-sni"] == healthcheck_tls_sni
        assert elastic_ip.healthcheck_tls_sni == healthcheck_tls_sni
        assert res["healthcheck"]["tls-skip-verify"] == healthcheck_tls_skip_verify
        assert elastic_ip.healthcheck_tls_skip_verify == healthcheck_tls_skip_verify

    def test_delete(self, exo, eip, instance):
        elastic_ip = ElasticIP._from_cs(exo.compute, eip(teardown=False))
        elastic_ip_id = elastic_ip.id
        instance = Instance._from_cs(exo.compute, instance())

        res = exo.compute.cs.listNics(virtualmachineid=instance.id, fetch_list=True)
        exo.compute.cs.addIpToNic(
            nicid=instance._default_nic(res)["id"], ipaddress=elastic_ip.address
        )

        elastic_ip.delete(detach_instances=True)
        assert elastic_ip.id is None

        res = exo.compute.cs.listPublicIpAddresses(id=elastic_ip_id, fetch_list=True)
        assert len(res) == 0

    def test_properties(self, exo, eip, instance, test_reverse_dns):
        elastic_ip = ElasticIP._from_cs(exo.compute, eip())
        instance = Instance._from_cs(exo.compute, instance())

        instance_nics = exo.compute.cs.listNics(
            virtualmachineid=instance.id, fetch_list=True
        )
        default_nic = instance._default_nic(instance_nics)
        exo.compute.cs.addIpToNic(nicid=default_nic["id"], ipaddress=elastic_ip.address)

        elastic_ip_instances = list(elastic_ip.instances)
        assert len(elastic_ip_instances) == 1
        assert elastic_ip_instances[0].name == instance.name

        exo.compute.cs.updateReverseDnsForPublicIpAddress(
            id=elastic_ip.id, domainname=test_reverse_dns
        )
        assert elastic_ip.reverse_dns == test_reverse_dns



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from base64 import b64decode
from cs import CloudStackApiException
from datetime import datetime, timedelta
from time import sleep
from exoscale.api import ResourceNotFoundError
from exoscale.api.compute import *
from .conftest import _random_str


class TestCompute:
    ### Anti-Affinity Group

    def test_create_anti_affinity_group(self, exo, test_prefix, test_description):
        anti_affinity_group_name = "-".join([test_prefix, _random_str()])
        anti_affinity_group = exo.compute.create_anti_affinity_group(
            name=anti_affinity_group_name, description=test_description
        )
        assert anti_affinity_group.id != ""
        assert anti_affinity_group.name == anti_affinity_group_name
        assert anti_affinity_group.description == test_description

        exo.compute.cs.deleteAffinityGroup(id=anti_affinity_group.id)

    def test_list_anti_affinity_groups(self, exo, aag):
        anti_affinity_group = AntiAffinityGroup._from_cs(exo.compute, aag())

        anti_affinity_groups = list(exo.compute.list_anti_affinity_groups())
        # We cannot guarantee that there will be only our resources in the
        # testing environment, so we ensure we get at least our fixture anti-affinity group
        assert len(anti_affinity_groups) >= 1

    def test_get_anti_affinity_group(self, exo, aag):
        anti_affinity_group1 = AntiAffinityGroup._from_cs(exo.compute, aag())
        anti_affinity_group2 = AntiAffinityGroup._from_cs(exo.compute, aag())

        anti_affinity_group = exo.compute.get_anti_affinity_group(
            id=anti_affinity_group1.id
        )
        assert anti_affinity_group.id == anti_affinity_group1.id

        anti_affinity_group = exo.compute.get_anti_affinity_group(
            name=anti_affinity_group2.name
        )
        assert anti_affinity_group.id == anti_affinity_group2.id

        with pytest.raises(ResourceNotFoundError) as excinfo:
            anti_affinity_group = exo.compute.get_anti_affinity_group(
                id="00000000-0000-0000-0000-000000000000"
            )
            assert anti_affinity_group is None
        assert excinfo.type == ResourceNotFoundError

        with pytest.raises(ResourceNotFoundError) as excinfo:
            anti_affinity_group = exo.compute.get_anti_affinity_group(name="lolnope")
            assert anti_affinity_group is None
        assert excinfo.type == ResourceNotFoundError

    ### Elastic IP

    def test_create_elastic_ip(self, exo, zone, test_description):
        zone = Zone._from_cs(zone("ch-gva-2"))
        healthcheck_mode = "https"
        healthcheck_port = 443
        healthcheck_path = "/health"
        healthcheck_interval = 5
        healthcheck_timeout = 3
        healthcheck_strikes_ok = 2
        healthcheck_strikes_fail = 1
        healthcheck_tls_sni = "example.net"
        healthcheck_tls_skip_verify = True

        elastic_ip = exo.compute.create_elastic_ip(
            zone=zone,
            description=test_description,
            healthcheck_mode=healthcheck_mode,
            healthcheck_port=healthcheck_port,
            healthcheck_path=healthcheck_path,
            healthcheck_interval=healthcheck_interval,
            healthcheck_timeout=healthcheck_timeout,
            healthcheck_strikes_ok=healthcheck_strikes_ok,
            healthcheck_strikes_fail=healthcheck_strikes_fail,
            healthcheck_tls_sni=healthcheck_tls_sni,
            healthcheck_tls_skip_verify=healthcheck_tls_skip_verify,
        )
        assert elastic_ip.zone.id == zone.id
        assert elastic_ip.zone.name == zone.name
        assert elastic_ip.address != ""
        assert elastic_ip.description == test_description
        assert elastic_ip.healthcheck_mode == healthcheck_mode
        assert elastic_ip.healthcheck_port == healthcheck_port
        assert elastic_ip.healthcheck_path == healthcheck_path
        assert elastic_ip.healthcheck_interval == healthcheck_interval
        assert elastic_ip.healthcheck_timeout == healthcheck_timeout
        assert elastic_ip.healthcheck_strikes_ok == healthcheck_strikes_ok
        assert elastic_ip.healthcheck_strikes_fail == healthcheck_strikes_fail
        assert elastic_ip.healthcheck_tls_sni == healthcheck_tls_sni
        assert elastic_ip.healthcheck_tls_skip_verify == healthcheck_tls_skip_verify

        exo.compute.cs.disassociateIpAddress(id=elastic_ip.id)

    def test_list_elastic_ips(self, exo, zone, eip):
        zone = Zone._from_cs(zone("ch-gva-2"))
        elastic_ip = eip(zone_id=zone.id)

        elastic_ips = list(exo.compute.list_elastic_ips(zone=zone))
        # We cannot guarantee that there will be only our resources in the
        # testing environment, so we ensure we get at least our fixture Elastic IP
        assert len(elastic_ips) >= 1

    def test_get_elastic_ip(self, exo, zone, eip, test_description):
        zone = Zone._from_cs(zone("ch-gva-2"))
        elastic_ip1 = ElasticIP._from_cs(exo.compute, eip())
        elastic_ip2 = ElasticIP._from_cs(exo.compute, eip())

        elastic_ip = exo.compute.get_elastic_ip(zone=zone, id=elastic_ip1.id)
        assert elastic_ip.id == elastic_ip1.id
        assert elastic_ip.description == test_description

        elastic_ip = exo.compute.get_elastic_ip(zone=zone, address=elastic_ip2.address)
        assert elastic_ip.id == elastic_ip2.id

        with pytest.raises(ResourceNotFoundError) as excinfo:
            elastic_ip = exo.compute.get_elastic_ip(
                zone=zone, id="00000000-0000-0000-0000-000000000000"
            )
            assert elastic_ip is None
        assert excinfo.type == ResourceNotFoundError

        with pytest.raises(ResourceNotFoundError) as excinfo:
            elastic_ip = exo.compute.get_elastic_ip(zone=zone, address="1.2.3.4")
            assert elastic_ip is None
        assert excinfo.type == ResourceNotFoundError

    ### Instance

    def test_create_instance(
        self,
        exo,
        zone,
        aag,
        sg,
        privnet,
        sshkey,
        instance_type,
        instance_template,
        test_prefix,
        test_instance_service_offering_id,
        test_instance_template_id,
    ):
        zone = Zone._from_cs(zone("ch-gva-2"))
        instance_type = InstanceType._from_cs(
            instance_type(id=test_instance_service_offering_id)
        )
        instance_template = InstanceTemplate._from_cs(
            exo.compute, instance_template(id=test_instance_template_id)
        )
        anti_affinity_group = AntiAffinityGroup._from_cs(exo.compute, aag())
        private_network = PrivateNetwork._from_cs(exo.compute, privnet())
        security_group1 = SecurityGroup._from_cs(exo.compute, sg())
        security_group2 = SecurityGroup._from_cs(exo.compute, sg())
        ssh_key = SSHKey._from_cs(exo.compute, sshkey())
        instance_name = "-".join([test_prefix, _random_str()])

        instance = exo.compute.create_instance(
            name=instance_name,
            zone=zone,
            type=instance_type,
            template=instance_template,
            volume_size=20,
            security_groups=[security_group1, security_group2],
            anti_affinity_groups=[anti_affinity_group],
            private_networks=[private_network],
            enable_ipv6=True,
            ssh_key=ssh_key,
        )
        assert instance.id != ""
        assert instance.name == instance_name
        assert datetime.now() - instance.creation_date.replace(tzinfo=None) < timedelta(
            minutes=2
        )
        assert instance.zone.id == zone.id
        assert instance.type.id == instance_type.id
        assert instance.template.id == instance_template.id
        assert instance.volume_size == 21474836480  # 20 GB
        assert instance.ipv4_address != ""
        assert instance.ipv6_address != ""
        assert instance.ssh_key.name == ssh_key.name
        assert (
            list(i["name"] for i in instance.res["securitygroup"]).sort()
            == [security_group1.name, security_group2.name].sort()
        )
        assert instance.res["affinitygroup"][0]["id"] == anti_affinity_group.id
        assert (
            list(i for i in instance.res["nic"] if not i["isdefault"])[0]["networkid"]
            == private_network.id
        )

        exo.compute.cs.destroyVirtualMachine(id=instance.id)

    def test_list_instances(self, exo, zone, privnet, instance):
        zone = Zone._from_cs(zone("ch-gva-2"))
        private_network = PrivateNetwork._from_cs(exo.compute, privnet())
        instance = Instance._from_cs(exo.compute, instance())

        instances = list(exo.compute.list_instances(zone=zone))
        # We cannot guarantee that there will be only our resources in the
        # testing environment, so we ensure we get at least our fixture instance
        assert len(instances) >= 1

        instances = list(exo.compute.list_instances(zone=zone, name=instance.name))
        assert len(instances) == 1

        res = exo.compute.cs.addNicToVirtualMachine(
            virtualmachineid=instance.id, networkid=private_network.id
        )

        instances = list(
            exo.compute.list_instances(zone=zone, networkid=private_network.id)
        )
        assert len(instances) == 1

        for nic in res["virtualmachine"]["nic"]:
            if nic["isdefault"]:
                continue
            exo.compute.cs.removeNicFromVirtualMachine(
                virtualmachineid=instance.id, nicid=nic["id"]
            )

    def test_get_instance(self, exo, zone, privnet, instance):
        zone = Zone._from_cs(zone("ch-gva-2"))
        private_network = PrivateNetwork._from_cs(exo.compute, privnet())
        instance = Instance._from_cs(exo.compute, instance())

        _instance = exo.compute.get_instance(zone=zone, id=instance.id)
        assert _instance.id == instance.id

        _instance = exo.compute.get_instance(
            zone=zone, ip_address=instance.ipv4_address
        )
        assert _instance.id == instance.id

        with pytest.raises(ResourceNotFoundError) as excinfo:
            _instance = exo.compute.get_instance(
                zone=zone, id="00000000-0000-0000-0000-000000000000"
            )
            assert _instance is None
        assert excinfo.type == ResourceNotFoundError

    ### Instance Template

    def test_list_instance_templates(
        self, exo, zone, test_instance_template_id, test_instance_template_name
    ):
        zone = Zone._from_cs(zone("ch-gva-2"))

        instance_templates = list(exo.compute.list_instance_templates(zone=zone))
        assert len(instance_templates) > 10

        instance_templates = list(
            exo.compute.list_instance_templates(
                zone=zone, name=test_instance_template_name
            )
        )
        assert len(instance_templates) == 1

    def test_get_instance_template(self, exo, zone, test_instance_template_id):
        zone = Zone._from_cs(zone("ch-gva-2"))

        instance_template = exo.compute.get_instance_template(
            zone=zone, id=test_instance_template_id
        )
        assert instance_template.id == test_instance_template_id

        with pytest.raises(ResourceNotFoundError) as excinfo:
            instance_template = exo.compute.get_instance_template(
                zone=zone, id="00000000-0000-0000-0000-000000000000"
            )
            assert instance_template is None
        assert excinfo.type == ResourceNotFoundError

    ### Instance Type

    def test_list_instance_types(self, exo):
        expected_type_name = "Micro"
        expected_type_id = "71004023-bb72-4a97-b1e9-bc66dfce9470"  # Micro
        expected_types = [
            "Micro",
            "Tiny",
            "Small",
            "Medium",
            "Large",
            "Extra-large",
            "Huge",
            "Mega",
            "Titan",
            "Jumbo",
        ]

        types = list(exo.compute.list_instance_types())
        assert len(types) > len(expected_types)

    def test_get_instance_type(self, exo):
        expected_type_name = "Micro"
        expected_type_id = "71004023-bb72-4a97-b1e9-bc66dfce9470"  # Micro

        instance_type = exo.compute.get_instance_type(name=expected_type_name)
        assert instance_type.name == expected_type_name
        assert instance_type.id == expected_type_id

        instance_type = exo.compute.get_instance_type(id=expected_type_id)
        assert instance_type.name == expected_type_name
        assert instance_type.id == expected_type_id

        with pytest.raises(ResourceNotFoundError) as excinfo:
            instance_type = exo.compute.get_instance_type(
                id="00000000-0000-0000-0000-000000000000"
            )
            assert instance_type is None
        assert excinfo.type == ResourceNotFoundError

        with pytest.raises(ResourceNotFoundError) as excinfo:
            instance_type = exo.compute.get_instance_type(name="lolnope")
            assert instance_type is None
        assert excinfo.type == ResourceNotFoundError

    ### Instance Pool

    def test_create_instance_pool(
        self,
        exo,
        zone,
        sg,
        privnet,
        sshkey,
        instance_type,
        instance_template,
        test_prefix,
        test_description,
        test_instance_service_offering_id,
        test_instance_template_id,
        test_instance_user_data,
    ):
        zone = Zone._from_cs(zone("ch-gva-2"))
        instance_pool_name = "-".join([test_prefix, _random_str()])
        instance_pool_size = 1
        instance_type = InstanceType._from_cs(instance_type())
        instance_template = InstanceTemplate._from_cs(
            exo.compute, instance_template(id=test_instance_template_id)
        )
        private_network = PrivateNetwork._from_cs(exo.compute, privnet(zone_id=zone.id))
        security_group = SecurityGroup._from_cs(exo.compute, sg())
        ssh_key = SSHKey._from_cs(exo.compute, sshkey())

        instance_pool = exo.compute.create_instance_pool(
            zone=zone,
            name=instance_pool_name,
            description=test_description,
            size=instance_pool_size,
            instance_type=instance_type,
            instance_template=instance_template,
            instance_volume_size=20,
            instance_security_groups=[security_group],
            instance_private_networks=[private_network],
            instance_ssh_key=ssh_key,
            instance_user_data=test_instance_user_data,
        )

        [actual_instance_pool] = exo.compute.cs.getInstancePool(
            id=instance_pool.id, zoneid=zone.id, fetch_list=True
        )

        assert instance_pool.zone.id == zone.id
        assert instance_pool.id == actual_instance_pool["id"]
        assert instance_pool.name == instance_pool_name
        assert actual_instance_pool["name"] == instance_pool_name
        assert instance_pool.description == test_description
        assert actual_instance_pool["description"] == test_description
        assert instance_pool.size == instance_pool_size
        assert actual_instance_pool["size"] == instance_pool_size
        assert instance_pool.instance_type.id == instance_type.id
        assert actual_instance_pool["serviceofferingid"] == instance_type.id
        assert instance_pool.instance_template.id == instance_template.id
        assert actual_instance_pool["templateid"] == instance_template.id
        assert instance_pool.instance_volume_size == 20
        assert actual_instance_pool["rootdisksize"] == 20
        assert (
            b64decode(instance_pool.instance_user_data).decode("utf-8")
            == test_instance_user_data
        )
        assert (
            b64decode(actual_instance_pool["userdata"]).decode("utf-8")
            == test_instance_user_data
        )

        exo.compute.cs.destroyInstancePool(id=instance_pool.id, zoneid=zone.id)

        # We have to delay the end of the test until the Instance Pool and its managed
        # Compute instances are effectively destroyed, otherwise the referenced fixtures
        # teardown will fail as it's not possible to delete resources still being
        # referenced by a Compute instance or Instance Pool.
        # Grab a seat, this can take a while...
        t = datetime.now()
        while (
            len(
                list(
                    i
                    for i in exo.compute.cs.listInstancePools(
                        zoneid=zone.id, fetch_list=True
                    )
                    if i["id"] == instance_pool.id
                )
            )
            > 0
        ) and datetime.now() - t < timedelta(minutes=5):
            sleep(10)

    def test_list_instance_pools(self, exo, zone, instance_pool):
        zone = Zone._from_cs(zone("ch-gva-2"))
        instance_pool = InstancePool._from_cs(
            exo.compute, instance_pool(zone_id=zone.id)
        )

        instance_pools = list(exo.compute.list_instance_pools(zone=zone))
        # We cannot guarantee that there will be only our resources in the
        # testing environment, so we ensure we get at least our fixture Instance Pool
        assert len(instance_pools) >= 1

    def test_get_instance_pool(self, exo, zone, instance_pool):
        zone = Zone._from_cs(zone("ch-gva-2"))
        instance_pool = InstancePool._from_cs(
            exo.compute, instance_pool(zone_id=zone.id)
        )

        _instance_pool = exo.compute.get_instance_pool(zone=zone, id=instance_pool.id)
        assert _instance_pool.id == instance_pool.id

        with pytest.raises(ResourceNotFoundError) as excinfo:
            _instance_pool = exo.compute.get_instance_pool(
                zone=zone, id="00000000-0000-0000-0000-000000000000"
            )
            assert _instance_pool is None
        assert excinfo.type == ResourceNotFoundError

    ### Network Load Balancer

    def test_create_network_load_balancer(
        self, exo, zone, test_prefix, test_description
    ):
        zone = Zone._from_cs(zone("ch-gva-2"))
        nlb_name = "-".join([test_prefix, _random_str()])

        nlb = exo.compute.create_network_load_balancer(
            zone=zone,
            name=nlb_name,
            description=test_description,
        )
        assert nlb.zone.id == zone.id
        assert nlb.zone.name == zone.name
        assert nlb.name == nlb_name
        assert nlb.description == test_description
        assert datetime.now() - nlb.creation_date.replace(tzinfo=None) < timedelta(
            minutes=2
        )

        res = exo.compute._v2_request_async(
            "DELETE", "/load-balancer/" + nlb.id, zone.name
        )

    def test_list_network_load_balancers(self, exo, zone, nlb):
        zone = Zone._from_cs(zone("ch-gva-2"))
        _ = nlb(zone_name=zone.name)

        nlbs = list(exo.compute.list_network_load_balancers(zone=zone))
        # We cannot guarantee that there will be only our resources in the
        # testing environment, so we ensure we get at least our fixture Network Load
        # Balancer
        assert len(nlbs) >= 1

    def test_get_network_load_balancer(self, exo, zone, nlb):
        zone = Zone._from_cs(zone("ch-gva-2"))
        nlb = NetworkLoadBalancer._from_api(exo.compute, nlb(), zone)

        _nlb = exo.compute.get_network_load_balancer(zone=zone, name=nlb.name)
        assert _nlb.name == nlb.name

        _nlb = exo.compute.get_network_load_balancer(zone=zone, id=nlb.id)
        assert _nlb.id == nlb.id

        with pytest.raises(ResourceNotFoundError) as excinfo:
            _nlb = exo.compute.get_network_load_balancer(
                zone=zone, id="00000000-0000-0000-0000-000000000000"
            )
            assert _nlb is None
        assert excinfo.type == ResourceNotFoundError

    ### Private Network

    def test_create_private_network(self, exo, zone, test_prefix, test_description):
        zone = Zone._from_cs(zone("ch-gva-2"))
        private_network_name = "-".join([test_prefix, _random_str()])
        start_ip = "192.168.1.10"
        end_ip = "192.168.1.100"
        netmask = "255.255.255.0"

        private_network = exo.compute.create_private_network(
            zone=zone,
            name=private_network_name,
            description=test_description,
            start_ip=start_ip,
            end_ip=end_ip,
            netmask=netmask,
        )
        assert private_network.zone.id == zone.id
        assert private_network.zone.name == zone.name
        assert private_network.name == private_network_name
        assert private_network.description == test_description
        assert private_network.start_ip == start_ip
        assert private_network.end_ip == end_ip
        assert private_network.netmask == netmask

        exo.compute.cs.deleteNetwork(id=private_network.id)

    def test_list_private_networks(self, exo, zone, privnet):
        zone = Zone._from_cs(zone("ch-gva-2"))
        private_network = PrivateNetwork._from_cs(exo.compute, privnet())

        private_networks = list(exo.compute.list_private_networks(zone=zone))
        # We cannot guarantee that there will be only our resources in the
        # testing environment, so we ensure we get at least our fixture Private Network
        assert len(private_networks) >= 1

        private_networks = list(
            exo.compute.list_private_networks(zone=zone, name=private_network.name)
        )
        assert len(private_networks) == 1
        assert private_networks[0].name == private_network.name

    def test_get_private_network(self, exo, zone, privnet):
        zone = Zone._from_cs(zone("ch-gva-2"))
        private_network = PrivateNetwork._from_cs(exo.compute, privnet())

        _private_network = exo.compute.get_private_network(
            zone=zone, id=private_network.id
        )
        assert _private_network.id == private_network.id

        with pytest.raises(ResourceNotFoundError) as excinfo:
            _private_network = exo.compute.get_private_network(
                zone=zone, id="00000000-0000-0000-0000-000000000000"
            )
            assert _private_network is None
        assert excinfo.type == ResourceNotFoundError

    ### Security Group

    def test_create_security_group(self, exo, test_prefix, test_description):
        security_group_name = "-".join([test_prefix, _random_str()])

        security_group = exo.compute.create_security_group(
            name=security_group_name, description=test_description
        )

        assert security_group.id != ""
        assert security_group.name == security_group_name
        assert security_group.description == test_description

        exo.compute.cs.deleteSecurityGroup(id=security_group.id)

    def test_list_security_groups(self, exo, sg):
        security_group1 = SecurityGroup._from_cs(exo.compute, sg())
        security_group2 = SecurityGroup._from_cs(exo.compute, sg())

        security_groups = list(exo.compute.list_security_groups())
        # We cannot guarantee that there will be only our resources in the
        # testing environment, so we ensure we get at least our 2 fixture security groups
        assert len(security_groups) >= 2

    def test_get_security_group(self, exo, sg):
        security_group1 = SecurityGroup._from_cs(exo.compute, sg())
        security_group2 = SecurityGroup._from_cs(exo.compute, sg())

        security_group = exo.compute.get_security_group(id=security_group1.id)
        assert security_group.name == security_group1.name
        assert security_group.description == security_group1.description

        security_group = exo.compute.get_security_group(name=security_group2.name)
        assert security_group.name == security_group2.name
        assert security_group.description == security_group2.description

        with pytest.raises(ResourceNotFoundError) as excinfo:
            security_group = exo.compute.get_security_group(
                id="00000000-0000-0000-0000-000000000000"
            )
            assert security_group is None
        assert excinfo.type == ResourceNotFoundError

        with pytest.raises(ResourceNotFoundError) as excinfo:
            security_group = exo.compute.get_security_group(name="lolnope")
            assert security_group is None
        assert excinfo.type == ResourceNotFoundError

    ### SSH Key

    def test_create_ssh_key(self, exo, test_prefix):
        ssh_key_name = "-".join([test_prefix, _random_str()])

        ssh_key = exo.compute.create_ssh_key(name=ssh_key_name)

        assert ssh_key.name == ssh_key_name
        assert ssh_key.fingerprint != ""
        assert ssh_key.private_key != ""

        exo.compute.cs.deleteSSHKeyPair(name=ssh_key.name)

    def test_register_ssh_key(self, exo, test_prefix):
        ssh_key_name = "-".join([test_prefix, _random_str()])

        ssh_key = exo.compute.register_ssh_key(
            name=ssh_key_name,
            public_key="ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDGRYWaNYBG/Ld3ZnXGsK9p"
            + "Zl9kT3B6GXvsslgy/LCjkJvDIP+nL+opAArKZD1P1+SGylCLt8ISdJNNGLtxKp9CL12EG"
            + "AYqdDvm5PurkpqIkEsfhsIG4dne9hNu7ZW8aHGHDWM62/4uiWOKtbGdv/P33L/Fepzypw"
            + "pivFsaXwPYVunAgoBQLUAmj/xcwtx7cvKS4zdj0+Iu21CIGU9wsH3ZLS34QiXtCGJyMOp"
            + "158qld9Oeus3Y/7DQ4w5XvfGn9sddxHOSMwUlNiFVty673X3exgMIc8psZOsHvWZPS0zW"
            + "x9gEDE95cUU10K6u4vzTr2O6fgDOQBynEUw3CDiHvwRD alice@example.net",
        )

        assert ssh_key.name == ssh_key_name
        assert ssh_key.fingerprint == "a0:25:fa:32:c0:18:7a:f8:e8:b2:3b:30:d8:ca:9a:2e"

        exo.compute.cs.deleteSSHKeyPair(name=ssh_key.name)

    def test_list_ssh_keys(self, exo, sshkey):
        ssh_key = SSHKey._from_cs(exo.compute, sshkey())

        ssh_keys = list(exo.compute.list_ssh_keys())
        # We cannot guarantee that there will be only our resources in the
        # testing environment, so we ensure we get at least our fixture SSH key
        assert len(ssh_keys) >= 1

    def test_get_ssh_key(self, exo, sshkey):
        ssh_key = SSHKey._from_cs(exo.compute, sshkey())

        _ssh_key = exo.compute.get_ssh_key(name=ssh_key.name)
        assert _ssh_key.name == ssh_key.name

        with pytest.raises(ResourceNotFoundError) as excinfo:
            _ssh_key = exo.compute.get_ssh_key(name="lolnope")
            assert _ssh_key is None
        assert excinfo.type == ResourceNotFoundError

    ### Zone

    def test_list_zones(self, exo):
        expected_zones = [
            "at-vie-1",
            "bg-sof-1",
            "ch-dk-2",
            "ch-gva-2",
            "de-fra-1",
            "de-muc-1",
        ]

        zones = list(exo.compute.list_zones())
        assert len(zones) >= len(expected_zones)

    def test_get_zone(self, exo):
        expected_zone_name = "ch-gva-2"
        expected_zone_id = "1128bd56-b4d9-4ac6-a7b9-c715b187ce11"

        zone = exo.compute.get_zone(id=expected_zone_id)
        assert zone.name == expected_zone_name
        assert zone.id == expected_zone_id

        zone = exo.compute.get_zone(name=expected_zone_name)
        assert zone.name == expected_zone_name
        assert zone.id == expected_zone_id



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pytest
from exoscale.api import ResourceNotFoundError
from exoscale.api.iam import *
from .conftest import _random_str


class TestIAM:
    ### API key

    def test_create_api_key(self, exo, test_prefix):
        api_key_name = "-".join([test_prefix, _random_str()])

        api_key = exo.iam.create_api_key(name=api_key_name)

        assert api_key.name == api_key_name
        assert api_key.key != ""
        assert api_key.operations != ""
        assert api_key.secret != ""
        assert api_key.type != ""

        exo.iam.cs.revokeApiKey(key=api_key.key)

    def test_list_api_key(self, exo, apikey):
        api_key = APIKey._from_cs(exo.iam, apikey())

        api_keys = list(exo.iam.list_api_keys())

        assert len(api_keys) >= 1

    def test_get_api_key(self, exo, apikey):
        api_key = APIKey._from_cs(exo.iam, apikey())

        _api_key = exo.iam.get_api_key(key=api_key.key)
        assert _api_key.name == api_key.name

        with pytest.raises(ResourceNotFoundError) as excinfo:
            _api_key = exo.iam.get_api_key(key="EXO000000000000000000000000")
            assert _api_key is None
        assert excinfo.type == ResourceNotFoundError

    def test_list_api_key_operations(self, exo):
        operations = list(exo.iam.list_api_key_operations())

        assert len(operations) >= 1



# -*- coding: utf-8 -*-

"""
Note:
    This module is not intended for standalone use, please use the :code:`exoscale`
    module or the :code:`exoscale.api.*` submodules.
"""


import attr
import pkg_resources
import platform
import requests
import sys
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from requests.utils import default_user_agent as requests_user_agent


@attr.s
class API(object):
    """
    An Exoscale API client.

    Attributes:
        endpoint (str): the API endpoint
        key (str): the API key
        secret (str): the API secret
        session (requests.Session): the API HTTP session
        max_retries (int): the API HTTP session retry policy number of retries to allow
        retry_backoff_factor (int): the API HTTP session retry policy backoff factor to
            apply between attempts after the second try
        trace (bool): API request/response tracing flag
    """

    endpoint = attr.ib()
    key = attr.ib()
    secret = attr.ib(repr=False)
    trace = attr.ib(default=False, repr=False)
    session = attr.ib(default=requests.Session(), repr=False)
    max_retries = attr.ib(default=3, repr=False)
    retry_backoff_factor = attr.ib(default=0.3, repr=False)

    def __attrs_post_init__(self):
        # HTTP session-related settings

        self.user_agent = (
            "Exoscale-Python/{python_exoscale_version} "
            "cs/{cs_version} "
            "{requests_user_agent} "
            "{python_implementation}/{python_version} "
            "{os_name}/{os_version}"
        ).format(
            python_exoscale_version=pkg_resources.get_distribution("exoscale").version,
            cs_version=pkg_resources.get_distribution("cs").version,
            requests_user_agent=requests_user_agent(),
            python_implementation=platform.python_implementation(),
            python_version=platform.python_version(),
            os_name=platform.system(),
            os_version=platform.release(),
        )

        adapter = HTTPAdapter(
            max_retries=Retry(
                total=self.max_retries,
                backoff_factor=self.retry_backoff_factor,
                status_forcelist=None,
            )
        )

        self.session.mount("http://", adapter=adapter)
        self.session.mount("https://", adapter=adapter)

    def send(self, **kwargs):
        """
        Send a signed HTTP request to the API.

        Parameters:
            **kwargs: requests.Request parameters

        Returns:
            request.Response: the HTTP request response
        """

        req = requests.Request(**kwargs).prepare()

        if self.trace:
            print(
                ">>> {method} {url}".format(method=req.method, url=req.url),
                file=sys.stderr,
            )
            print("    headers:{headers}".format(headers=req.headers), file=sys.stderr)
            if req.body:
                print("    body:{body}".format(body=req.body), file=sys.stderr)

        res = self.session.send(req)

        if self.trace:
            print(
                "<<< {status} {reason}".format(
                    status=res.status_code, reason=res.reason
                ),
                file=sys.stderr,
            )
            print("    headers:{headers}".format(headers=res.headers), file=sys.stderr)
            print("    body:{body}".format(body=res.text), file=sys.stderr)

        return res


@attr.s
class Resource(object):
    """
    A resource returned by the Exoscale API.
    """

    res = attr.ib(repr=False)


@attr.s
class APIException(Exception):
    """
    A generic API error.
    """

    reason = attr.ib()
    error = attr.ib(default=None, repr=False)


@attr.s
class RequestError(Exception):
    """
    An API request error.
    """

    reason = attr.ib()
    error = attr.ib(default=None, repr=False)


class ResourceNotFoundError(APIException):
    """
    An error indicating that requested resource cannot be found.
    """

    def __init__(self):
        super().__init__("resource not found")

    def __repr__(self):
        return self.reason

    def __str__(self):
        return self.__repr__()
